/**
 * @file cs2ccmb.cs 
 * Determines a CCMB model of a CCM component written in C#.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */
using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;

namespace CCM {

/**
 * cs2ccmb main program. Controls command line argument processing, parser
 * invocation and model generation.
 */
class CS2CCMB {
#if DEBUG
#else
    /// Number of nodes when #nodes was printed last
    static int lastPrintNodes = -10000;
#endif

    /// Calculate process for given state node.
    private static void calcProcess( StateNode s ) {
#if DEBUG
	Console.WriteLine( "calculating process for " + s.stateName() );
#endif
	Node branchNode;

	// call each [Event] Method
	foreach( Method m in eventMethods ) {
	    foreach( List args in m ) {
#if DEBUG
		Console.WriteLine( "calling [event] Method {0}({1})", m.getName(), args );
#endif
		branchNode = new MethodCallNode( null, m.getName(), args );
		s.addBranch( branchNode );
		StubIterator.start();
		do {
		    model.comp.setState( s );
		    Node.currentNode = branchNode;
		    m.execute( args );
		    runMethod.execute( emptyArgs );
		    model.terminateThread();
#if DEBUG
#else
		    if( Node.numNodes - lastPrintNodes > 1000 ) {
			Console.Write( Node.numNodes + "\r" );
			lastPrintNodes = Node.numNodes;
		    }

#endif
		} while( StubIterator.next() );
	    }
	} 

	StubIterator.start();
	do {
	    // call Run() on its own
	    model.comp.setState( s );
	    Node.currentNode = s;
	    runMethod.execute( emptyArgs );
	    model.terminateThread();
#if DEBUG
#else
	    if( Node.numNodes - lastPrintNodes > 1000 ) {
		Console.Write( Node.numNodes + "\r" );
		lastPrintNodes = Node.numNodes;
	    }

#endif
	} while( StubIterator.next() );
    }

    /**
     * Add CCM and C# standard classes to the given list of known classes. This
     * adds: 
     * - class Exception with ctor Exception( Expr )
     * - Common with Common.log( Object, string )
     */
    public static void addStandardClasses( Hashtable classes ) {
	// add Exception class
	Class cls = new Class( "Exception", null, "" );

	IList pars = new List(0);
	Block body = new Block( null );
	Method m = new Method( "Exception", null, pars, null, false );
	m.setBody( body );
	cls.addConstructor( m );

	pars = new List(1);
	pars.Add( new VariableDecl( "msg", "".GetType(), null, false, null, false ) );
	m = new Method( "Exception", null, pars, null, false );
	m.setBody( body);
	cls.addConstructor( m );

	classes.Add( cls.getName(), cls );

	// add Common class
	cls = new Class( "Common", null, "" );
	
	pars = new List(2);
	pars.Add( new VariableDecl( "originator", (new Object()).GetType(), null, false, null, false ) );
	pars.Add( new VariableDecl( "msg", "".GetType(), null, false, null, false ) );
	m = new Method( "log", null, pars, null, true );
	body = new Block( null );
	Expr write = new StringExpr( new Ref( new QID( "msg" ), false ) );
	body.addStm( new WriteStm( write, true ) );
	m.setBody( body );
	cls.addMethod( m );

	classes.Add( cls.getName(), cls );
    }

    /**
     * Parse output mode and filename into outputMode and outputFile.
     * @param args Array of command line arguments
     * @param idx Start at array position idx
     * @return Index of the first argument in args after the output mode
     *         or -1 on parsing error
     */
    public static int parseOutput( string[] args, int idx ) {
	if( args[idx] == "-ccmb" )
	    outputMode = OutputMode.CCMB;
	else if( args[idx] == "-prolog" )
	    outputMode = OutputMode.PROLOG;
	else if( args[idx] == "-bin" )
	    outputMode = OutputMode.BIN;
	else {
	    Console.WriteLine( "Error: invalid output mode (must be -ccmb, -prolog, or -bin)" );
	    return -1;
	}
	
	outputFile = args[idx+1];
	if( outputMode == OutputMode.BIN && outputFile == "-" ) {
	    Console.WriteLine( "Error: cannot output binary processes to standard output" );
	    return -1;
	}

	return idx+2;
    }

    /**
     * Parse all files given on command line and write the Comp class into
     * model.comp. 
     * @param args Array of command line arguments
     * @param idx Start at array position idx
     * @return Index of the first ctor argument in args (just after
     *         the -- argument) or -1 on parsing error
     */
    public static int parseAll( string[] args, int idx ) {
	int arg;
	TextReader reader;
	CSharpParser parser;
	Hashtable classes = new Hashtable();

	addStandardClasses( classes );

	for( arg = idx; arg < args.Length && args[arg] != "--"; ++arg ) {
	    reader = File.OpenText( args[arg] );
	    parser = new CSharpParser( classes );
	    if( !parser.parse( reader, args[arg] ) )
		return -1;
	}

	// find Comp among classes
	model.comp = null;
	foreach( DictionaryEntry i in classes ) {
	    Class c = (Class) i.Value;
	    if( c.parentClass() == "Comp" ) {
		if( model.comp == null )
		    model.comp = c;
		else {
		    Console.WriteLine( "Error: more than one Comp class" );
		    return -1;
		}
	    }
	}
	if( model.comp == null ) {
	    Console.WriteLine( "Error: no Comp class" );
	    return -1;
	}
	model.comp.becomeCurrentClass();

#if DEBUG
	Console.WriteLine( "--- Successfully parsed classes: --------" );
	foreach( DictionaryEntry i in classes ) 
	    Console.WriteLine( i.Value );
	Console.WriteLine( "Comp class: " + model.comp.getName() );
	Console.WriteLine();
#endif

	Ref.setClasses( classes );
	return arg+1;

    }

    /**
     * Parse all constructor arguments given on command line.
     * @param args Array of command line arguments
     * @param idx Start at array position idx
     * @return List of parsed arguments
     */
    public static IList parseCtorArgs( string[] args, int idx ) {
	Object o;
	IList result = new ArrayList();
	for(; idx < args.Length; ++idx ) {
	    try {
		o = bool.Parse( args[idx] );
	    } catch( FormatException ) {
		try {
		    o = int.Parse( args[idx] );
		} catch( FormatException ) {
		    try {
			o = double.Parse( args[idx] );
		    } catch( FormatException ) {
			o = args[idx]; // fall back to string
		    }
		}
	    }
	    result.Add( o );
	}
	return result;
    }

    /// Main method.
    public static int Main( string[] args )
    {
	if( args.Length == 0 ) {
	    Console.WriteLine( 
		    "Usage: cs2ccmb -ccmb|-prolog|-bin <outfile> <file.cs> [<file.cs> ...] [ -- <ctorarg> <ctorarg> ... ]" );
	    return 0;
	}

	// parse command line arguments

	int idx = parseOutput( args, 0 );
	if( idx == -1 ) {
	    Console.WriteLine( "Aborting." );
	    return -1;
	}

	idx = parseAll( args, idx );
	if( idx == -1 ) {
	    Console.WriteLine( "Aborting." );
	    return -1;
	}

	IList ctorArgs = parseCtorArgs( args, idx );

	// check sanity
	runMethod = model.comp.getMethod( "Run", emptyArgs ); 
	if( runMethod == null ) {
	    Console.WriteLine( "Error: class does not contain a method 'void Run()'" );
	    return -2;
	}

	// save start state
	model.setStartState();
	Node.currentNode = model.getStartState();

	eventMethods = model.comp.attrMethods( "Event" );

#if DEBUG
#else
	Console.WriteLine( "Number of calculated transitions:" );
#endif
	// call constructor and Run() 
	StubIterator.start();
	do {
	    model.comp.setState( model.getStartState() );
	    Node.currentNode = model.getStartState();
	    model.comp.construct( ctorArgs );
	    runMethod.execute( emptyArgs );
	    model.terminateThread();
#if DEBUG
#else
	    if( Node.numNodes - lastPrintNodes > 1000 ) {
		Console.Write( Node.numNodes + "\r" );
		lastPrintNodes = Node.numNodes;
	    }
#endif
	} while( StubIterator.next() );

#if DEBUG
	Console.WriteLine( "--- Class after constructor call: --------" );
	model.comp.dumpState();
	Console.WriteLine();

	Console.WriteLine( "--- Processes after constructor call: --------" );
	model.writeCCMB( Console.Out );
#endif

        // create all state processes
	StateNode todo;
        while( ( todo = model.nextTodo() ) != null  ) {
#if DEBUG
	    Console.WriteLine( "DEQUEUE todo " + todo.stateName() );
#endif
            calcProcess( todo );
#if DEBUG
	    Console.WriteLine( "--- Processes after calcProcess() : ------------" );
	    model.writeCCMB( Console.Out );
#endif
        }

#if DEBUG
	Console.WriteLine( "--- Processes : -------------------------" );
	model.writeCCMB( Console.Out );

	Console.WriteLine( "--- unresolved state abstraction: ------------------" );
	Console.WriteLine( model.comp.getStateAbstraction() );

	Console.WriteLine( "--- transition abstraction: ------------------------" );
	Console.WriteLine( model.comp.getTransitionAbstraction() );
#endif

	Console.WriteLine( "Finished. Calculated " + Node.numNodes + " nodes." );

	// write output
	if( outputMode == OutputMode.BIN ) {
	    try {
		FileStream fs = new FileStream( outputFile, FileMode.Create );
		model.writeBin( fs );
		fs.Close();
	    } catch( SerializationException e ) {
		Console.WriteLine( "Error: could not serialize processes: " + e );
		return -4;
	    } catch( Exception e ) {
		Console.WriteLine( "Error: could not open output file " + outputFile + ":\n" + e );
		return -3;
	    }
	} else {
	    TextWriter output;
	    try {
		if( outputFile == "-" )
		    output = Console.Out;
		else
		    output = new StreamWriter( outputFile );
	    } catch( Exception e ) {
		Console.WriteLine( "Error: could not open output file " + outputFile + ":\n" + e );
		return -3;
	    }

	    if( outputMode == OutputMode.CCMB )
		model.writeCCMB( output ); 
	    else
		model.writeProlog( output );
	    output.Close();
	}
	return 0;
    }

    /// List of all Methods with [Event] attribute.
    private static IList eventMethods;

    /// Constant empty list (used to call methods with no parameters).
    private static ArrayList emptyArgs = new ArrayList();

    /// Generated model.
    private static Model model = new Model();

    private static Method runMethod;

    private enum OutputMode { CCMB, PROLOG, BIN };
    private static OutputMode outputMode;
    private static string outputFile;
}

} // closing namespace
