/**
 * @file pairset.cs 
 * Efficient implementation for a set of object pairs.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.Collections;

namespace Sets {

/// Pair of objects.
public struct Pair {
    /// Create pair (l,r).
    public Pair( Object l, Object r ) {
	left = l;
	right = r;
    }

    public override string ToString() {
	return "(" + left + ", " + right + ")";
    }

    public Object left, right;
}

/**
 * Set for storing pairs of objects. Implemented with two cascading hashtables
 * for O(log(size)) performance. Implements IEnumerable to enable iteration.
 */
public class PairSet : IEnumerable {
    /// Add object pair (o1,o2) to the set.
    public void add( Object o1, Object o2 ) {
	Hashtable right = (Hashtable) elems[o1];
	if( right == null )
	    elems[o1] = right = new Hashtable();
	right[o2] = null;
    }

    /// Add given object pair to the set. Attention: the pair itself is not
    /// stored, so do not rely on Pair object identity on iteration.
    public void add( Pair p ) {
	add( p.left, p.right );
    }

    /// Remove object pair (o1,o2) from the set.
    public void remove( Object o1, Object o2 ) {
	Hashtable right = (Hashtable) elems[o1];
	if( right == null )
	    return;
	right.Remove( o2 );

	// clean up
	if( right.Count == 0 )
	    elems.Remove( o1 );
    }

    /// Remove object pair from the set.
    public void remove( Pair p ) {
	remove( p.left, p.right );
    }

    /// Remove all elements whose left part equals to o.
    public void removeWhereLeft( Object o ) {
	elems.Remove( o );
    }

    /// Remove all elements whose right part equals to o.
    public void removeWhereRight( Object o ) {
	IList removedL = new ArrayList();
	Hashtable h;
	foreach( DictionaryEntry i in elems ) {
	    h = ( (Hashtable) i.Value );
	    h.Remove( o );
	    if( h.Count == 0 )
		removedL.Add( i.Key );
	}
	foreach( Object i in removedL )
	    elems.Remove( i );
    }

    /// Return whether given object pair (o1,o2) is contained in the set.
    public bool contains( Object o1, Object o2 ) {
	Hashtable right = (Hashtable) elems[o1];
	if( right == null )
	    return false;
	return right.ContainsKey( o2 );
    }

    /// Return whether given object pair is contained in the set.
    public bool contains( Pair p ) {
	return contains( p.left, p.right );
    }

    /// Number of elements in the set.
    public int Count {
	get {
	    int count = 0;
	    foreach( DictionaryEntry i in elems )
		count += ( (Hashtable) i.Value ).Count;
	    return count;
	}
    }

    /// Return whether set is empty.
    public bool isEmpty() {
	return elems.Count == 0;
    }

    /// Choose an element, remove it from the set and return it. Do not use on
    /// empty sets (this returns a (null,null) pair)!
    public Pair take() {
	Pair p;
	IEnumerator i = GetEnumerator();
	if( i.MoveNext() ) {
	    p = (Pair) i.Current;
	    remove( p );
	    return p;
	}
	return new Pair( null, null );
    }

    /// Return an enumerator over all possible values.
    public IEnumerator GetEnumerator() {
	return new PairSetEnumerator( elems );
    }

    /// Return the set in mathematical syntax.
    public override string ToString() {
	string result = "{";
	bool first = true;
	foreach( Pair p in this ) {
	    if( first ) first = false;
	    else result += ", ";
	    result += p;
	}
	return result + "}";
    }

    /**
     * Stores the elements in this set in two cascaded hashtables (Object ->
     * Object -> null).
     */
    private Hashtable elems = new Hashtable();
}

/// Enumerator for PairSet.
public class PairSetEnumerator : IEnumerator {
    public PairSetEnumerator( Hashtable elems ) {
	left = elems.GetEnumerator();
	right = null;
    }

    public void Reset() {
	left.Reset();
	right = null;
    }

    public bool MoveNext() {
	if( right == null ) {
	    if( !left.MoveNext() )
		return false;
	    right = ( (Hashtable) left.Value ).GetEnumerator();
	}
	if( right.MoveNext() )
	    return true;
	else {
	    right = null;
	    return MoveNext();
	}
    }

    /// Return the current enumeration value as Pair.
    public Object Current {
	get { 
	    if( right == null ) 
		return null;
	    return new Pair( left.Key, right.Key );
	}
    }

    private IDictionaryEnumerator left, right;
}

}
