/**
 * @file sim.cs 
 * Reads two binary process dumps and check whether they are (bi)similar.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */
using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using Sets;

namespace CCM {

/**
 * (Bi)similarity checker. Compares two process models (given in binary files
 * created by cs2ccmb) and checks whether the first one is a consistent
 * abstraction of the second. The mode of operation (similarity or
 * bisimilarity) is determined by the field 'bisim'.
 */
public class SimChecker {
    /// Load (deserialize) model from given binary file. On failure, the error
    /// is printed and null is returned.
    public static Model loadModel( string filename ) {
	try {
	    FileStream fs = new FileStream( filename, FileMode.Open );
	    Model m = new Model( fs );
	    fs.Close();
	    return m;
	}
	catch( SerializationException e ) {
	    Console.WriteLine( "Error: could not deserialize processes: " + e );
	} catch( IOException e ) {
	    Console.WriteLine( "Error: could not open input file " + filename + ":" + e );
	}
	return null;
    }

    /// Main.
    public static int Main( string[] args ) {
	// parse options
	int argind;
	for( argind = 0; argind < args.Length; ++argind ) {
	    if( args[argind] == "-b" )
		bisim = true;
	    else if( args[argind] == "-d" )
		debug = true;
	    else
		break;
	}

	if( args.Length-argind < 2 ) {
	    Console.WriteLine( "Usage: sim.exe [-b] [-d] <abstract binfile> <impl binfile> [<condition> ...]\n" ); 
	    Console.WriteLine( "  -b\tbisimulation checking mode (default is implementation checking)" );
	    Console.WriteLine( "  -d\tverbose checking (for debugging)" );
	    return 0;
	}

	abstr = loadModel( args[argind] );
	Console.WriteLine( "Loaded abstract model." );
	impl = loadModel( args[argind+1] );
	Console.WriteLine( "Loaded implementation model." );

	if( abstr == null || impl == null )
	    return -1;

	abstr.stateAbstraction.resolve( impl.stateFormat );

	implPred = new StatePredicate( impl, args, argind+2 );

	if( debug ) {
	    Console.WriteLine( "Abstract state format: " + abstr.stateFormat );
	    Console.WriteLine( "Implementation state format: " + impl.stateFormat );
	    Console.WriteLine( abstr.stateAbstraction );
	    Console.WriteLine( abstr.stateAbstraction );
	    Console.WriteLine( abstr.transitionAbstraction );
	    Console.WriteLine( "Implementation state filter: " + implPred );
	}

	if( similar( abstr.getStartState(), impl.getStartState() ) )
	    Console.WriteLine( bisim ? "Models are bisimilar." : 
		    "First model is implemented by the second one." );
	else
	    Console.WriteLine( bisim ? "\n\nModels are not bisimilar." :
		    "\n\nFirst model is not implemented by the second one." );

	return 0;
    }

    /**
     * Search an i-successor that matches asucc. If a match is found, a new
     * edge is put into edges and true is returned; otherwise returns false.
     * @param a	Node that abstracts node i
     * @param asucc Successor node of a for which a corresponding
     *        i-successor is to be searched
     * @param i Node that implements node a
     */
    public static bool searchImpl( Node a, Node i, Node asucc ) {
	if( ++numNodes - lastPrintedNodes >= 1000 ) {
	    lastPrintedNodes = numNodes;
	    Console.Write( numNodes + "\r" );
	}

	if( debug ) log.log( "searchImpl: finding next pair to <{0},{1}> with <{2},X>", a, i, asucc );
	if( i.hasSuccessors() )
	    foreach( Node isucc in i ) {
		if( similar( asucc, isucc ) ) {
		    if( debug ) log.log( "searchImpl: found {0} -> true", isucc );
		    edges.add( new Pair( asucc, isucc ), new Pair( a, i ) );
		    return true;
		}
	    }
	else
	    if( debug ) log.log( "searchImpl: no successors" );
	if( debug ) log.log( "searchImpl: false" );
	return false;
    }

    /**
     * Search an a-successor that matches isucc. If a match is found, a new
     * edge is put into edges and true is returned; otherwise returns false.
     * @param a	Node that abstracts node i
     * @param i Node that implements node a
     * @param isucc Successor node of i for which a corresponding
     *        a-successor is to be searched
     */
    public static bool searchAbstr( Node a, Node i, Node isucc ) {
	if( debug ) log.log( "searchAbstr: finding next pair to <{0},{1}> with <X,{2}>", a, i, isucc );
	if( a.hasSuccessors() )
	    foreach( Node asucc in a ) {
		if( similar( asucc, isucc ) ) {
		    if( debug ) log.log( "searchAbstr: found {0} -> true", asucc );
		    edges.add( new Pair( asucc, isucc ), new Pair( a, i )  );
		    return true;
		}
	    }
	else
	    if( debug) log.log( "searchAbstr: no successors" );
	if( debug ) log.log( "searchImpl: false" );
	return false;
    }

    /**
     * Check whether given nodes are similar wrt. to the given abstraction
     * relations. If this.bisim==true, then it checks for bisimilarity.
     * @param anode Abstract node
     * @param inode Implementation node
     */
    public static bool similar( Node anode, Node inode ) {
	if( debug ) {
	    log.log( "Checking {0} ~ {1} ?", anode, inode );
	    ++log;
	}

	// resolve state references
	if( anode is StateRefNode )
	    anode = (StateNode) abstr.getStates()[anode.ToString()];
	if( inode is StateRefNode )
	    inode = (StateNode) impl.getStates()[inode.ToString()];

	if( anode.GetType() != inode.GetType() ) {
	    if( debug ) {
		log.log( "Nodes have different type, returning false" );
		--log;
	    }
	    return false;
	}

	// already evaluated and found to be unrelated
	if( unrelated.contains( anode, inode ) ) {
	    if( debug ) {
		log.log( "Nodes already known as unrelated, returning false" );
		--log;
	    }
	    return false;
	}

	if( inode is StateNode ) { 
	    // filtered out?
	    if( !implPred.contains( (StateNode) inode ) ) {
		if( debug ) {
		    log.log( "Implementation state is not contained in filter predicate -> related" );
		    --log;
		}
		return true;
	    }

	    // state nodes match the state abstraction?
	    if( !abstr.stateAbstraction.equal( (StateNode) anode, 
			(StateNode) inode) ) {
		if( debug ) {
		    log.log( "States do not match, adding to unrelated" );
		    --log;
		}
		unrelated.add( anode, inode );
		return false;
	    }
	} else {
	    // transitions match the transition abstraction?
	    if( !abstr.transitionAbstraction.equal( (TransitionNode) anode,
			(TransitionNode) inode ) ) {
		if( debug ) {
		    log.log( "Transitions do not match, adding to unrelated" );
		    --log;
		}
		unrelated.add( anode, inode );
		return false;
	    }
	}

	// already evaluated and found to be related
	if( vert.contains( anode, inode ) ) {
	    if( debug ) {
		log.log( "Nodes already known related, returning true" );
		--log;
	    }
	    return true;
	}

	// <anode,inode> is analyzed for the first time
	vert.add( anode, inode );
	bool status = true;

	// match each successor of anode to a successor of inode
	if( anode.hasSuccessors() )
	    foreach( Node asucc in anode ) {
		status = searchImpl( anode, inode, asucc );
		if( !status )
		    break;
	    }

	if( debug ) log.log( "Status after searching matching impl successors: {0}", status );

	// match each successor of inode to a successor of anode when in
	// bisimulation mode
	if( bisim && status && inode.hasSuccessors() ) {
	    foreach( Node isucc in inode ) {
		status = searchAbstr( anode, inode, isucc );
		if( !status )
		    break;
	    }

	    if( debug ) log.log( "Status after searching matching abstract successors: {0}", status );
	}

	if( !status ) {
	    PairSet toremove = new PairSet();
	    toremove.add( anode, inode );

	    if( !debug ) {
		Console.WriteLine( "----------------------------------------------------");
		if( bisim )
		    Console.Write( "Not bi" );
		else
		    Console.Write( "Not forward " );
		Console.WriteLine( "similar:\nAbstract      : {0}\nImplementation: {1}",
		    anode, inode );
	    }

	    // recursively mark parent vertices as unrelated
	    Pair rmnode;
	    while( !toremove.isEmpty() ) {
		if( debug ) log.log( "to be removed: " + toremove );
		rmnode = toremove.take();
		if( debug ) log.log( "Marking state " + rmnode + " as unrelated " );

		// add predecessor nodes of rmode to affected
		foreach( Pair edge in edges ) {
		    if( edge.left.Equals( rmnode ) )
			toremove.add( (Pair) edge.right );
		}

		// remove rmnode and edges from and to it
		edges.removeWhereLeft( rmnode );
		edges.removeWhereRight( rmnode );
		vert.remove( rmnode );

		// mark as unrelated
		unrelated.add( rmnode );
	    }
	} 

	if( debug ) {
	    log.log( "-> {0} ~ {1} = {2}", anode, inode, status );
	    log.log( "Related now: " + vert );
	    --log;
	}
	return status;
    }

    /// Abstract model.
    private static Model abstr;
    
    /// Implementation model.
    private static Model impl;

    /// Predicate for implementation states (states not contained in this
    /// do not need to satisfy bisimulation requirements).
    private static StatePredicate implPred;

    /// Vertices of the bisimulation graph (contains (Node,Node) ).
    private static PairSet vert = new PairSet();

    /// Edges of the bisimulation graph; contains predecessor pairs
    /// (Pair',Pair), where the pairs are vertices (Node,Node).
    private static PairSet edges = new PairSet();

    /// Set of node pairs that are known to be unrelated.
    private static PairSet unrelated = new PairSet();

    private static bool bisim = false;

    /// Logger for debugging output.
    private static Debugging.Logger log = new Debugging.Logger();

    /// Flag wheter checking is verbose (for debugging).
    private static bool debug = false;

    private static int numNodes = 0, lastPrintedNodes = 0;
}

}
