/** 
 * @file statepredicate.cs 
 * Classes to implement prediates (filters) over process states. This allows
 * checking (bi)simulation of an abstraction agains only some part of an
 * implementation.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.Collections;
using System.Text.RegularExpressions;

namespace CCM {

/// Abstract value for use in predicate conditions.
abstract class ConditionValue {
    public abstract Object eval( StateNode state );
}

/// Boolean constant for use in predicate conditions.
class ConditionBoolValue : ConditionValue {
    public ConditionBoolValue( bool v ) {
	value = v;
    }

    public override Object eval( StateNode state ) {
	return value;
    }

    public override string ToString() {
	return value.ToString();
    }

    bool value;
}

/// Integer constant for use in predicate conditions.
class ConditionIntValue : ConditionValue {
    public ConditionIntValue( int v ) {
	value = v;
    }

    public override Object eval( StateNode state ) {
	return value;
    }

    public override string ToString() {
	return value.ToString();
    }

    int value;
}

/// Double constant for use in predicate conditions.
class ConditionDoubleValue : ConditionValue {
    public ConditionDoubleValue( double v ) {
	value = v;
    }

    public override Object eval( StateNode state ) {
	return value;
    }

    public override string ToString() {
	return value.ToString();
    }

    double value;
}

/// State field value for use in predicate conditions.
class ConditionFieldValue : ConditionValue {
    public ConditionFieldValue( string name, int index ) {
	this.name = name;
	this.index = index;
    }

    public override Object eval( StateNode state ) {
	return state.getState()[index];
    }

    public override string ToString() {
	return name + "(index " + index + ")";
    }

    int index;
    string name;
}

/// Single predicate condition.
class PredCondition {
    /**
     * Initialize single predicate condition.
     * @see StatePredicate.StatePredicate
     */
    public PredCondition( Model model, string spec ) {
	Regex specformat = new Regex( "^\\s*(-?(?:\\w|\\.)*)\\s*(<=|<|=|>|>=)\\s*(-?(?:\\w|\\.)*)\\s*$" );
	Match m = specformat.Match( spec );
	if( m.Success ) {
	    left = parseValue( model, m.Groups[1].Value );
	    right = parseValue( model, m.Groups[3].Value );

	    if( m.Groups[2].Value == "<" )
		op = Op.LT;
	    else if( m.Groups[2].Value == "<=" )
		op = Op.LE;
	    else if( m.Groups[2].Value == "=" )
		op = Op.EQ;
	    else if( m.Groups[2].Value == ">=" )
		op = Op.GE;
	    else
		op = Op.GT;
	} else
	    throw new Exception( "Invalid predicate condition format. Must be 'value <|<=|=|>=|> value'" );

    }

    /// Returns whether the given state node satisfies this condition.
    public bool contains( StateNode state ) {
	Object lv = left.eval( state );
	Object rv = right.eval( state );
	if( op == Op.EQ )
	    return lv.Equals( rv );
	if( lv.GetType() != rv.GetType() )
	    throw new Exception( "Values " + left + " and " + right + " have different data types" );

	if( lv is int ) {
	    int li = (int) lv;
	    int ri = (int) rv;

	    switch( op ) {
		case Op.LT: return li < ri;
		case Op.LE: return li <= ri;
		case Op.GE: return li >= ri;
		case Op.GT: return li > ri;
	    }
	} else if( lv is double ) {
	    double ld = (double) lv;
	    double rd = (double) rv;

	    switch( op ) {
		case Op.LT: return ld < rd;
		case Op.LE: return ld <= rd;
		case Op.GE: return ld >= rd;
		case Op.GT: return ld > rd;
	    }
	} else
	    throw new Exception( "Value " + left + " must be integer or double type" );

	// not reached
	return false;
    }

    public override string ToString() {
	string ops;
	switch( op ) {
	    case Op.LT: ops = "<"; break;
	    case Op.LE: ops = "<="; break;
	    case Op.EQ: ops = "="; break;
	    case Op.GT: ops = ">"; break;
	    default: ops = ">="; break;
	}

	return left + " " + ops + " " + right;
    }

    private static ConditionValue parseValue( Model model, string val ) {
	try {
	    return new ConditionBoolValue( bool.Parse( val ) );
	} catch( FormatException ) {
            try {
                return new ConditionIntValue( int.Parse( val ) );
            } catch( FormatException ) {
                try {
                    return new ConditionDoubleValue( double.Parse( val ) );
                } catch( FormatException ) {
		    int idx = model.fieldIndex( val );
		    if( idx >= 0 )
			return new ConditionFieldValue( val, idx );
		    throw new Exception( "Invalid value " + val );
		}
	    }
	}
    }

    enum Op { LT, LE, EQ, GE, GT };
    Op op;
    ConditionValue left, right;
}

/// Predicate over a StateNode.
public class StatePredicate {
    /**
     * Initializes the state node predicate for the given model. spec is a
     * string array of conditions of the form "value op value" (starting at
     * index specoffset). op is a binary relation <, <=, =, >=, or >, and value
     * is either a constant or a state field name. All conditions must be
     * satisfied to satisfy the state predicate.
     */
    public StatePredicate( Model model, string[] spec, int specoffset ) {
	fieldPreds = new List();
	for( int i = specoffset; i < spec.Length; ++i )
	    fieldPreds.Add( new PredCondition( model, spec[i] ) );
    }

    /// Returns whether the given state node satisfies all conditions.
    public bool contains( StateNode state ) {
	foreach( PredCondition fp in fieldPreds )
	    if( !fp.contains( state ) )
		return false;
	return true;
    }

    public override string ToString() {
	return fieldPreds.ToString();
    }

    private IList fieldPreds;
}

}

