/**
 * @file cs-parser.cs
 * Parse a C# file into a representing object tree.
 *
 * This file is based on cs-parser.jay of the Mono project (www.go-mono.com).
 * Changed by Martin Pitt <martin@piware.de> for a diploma thesis: I removed
 * the original C# code completely and only use the grammar rules.
 *
 * Original grammar authors:
 *    Miguel de Icaza (miguel@gnu.org)
 *    Ravi Pratap     (ravi@ximian.com)
 *
 * Code author;
 *    Martin Pitt     (martin@piware.de)
 *
 * Licensed under the terms of the GNU GPL
 *
 * Portions (C) 2001 Ximian, Inc (http://www.ximian.com)
 * Portions (C) 2004 Martin Pitt <martin@piware.de>
 */

using System;
using System.Text;
using System.IO;
using System.Collections;

namespace CCM
{
	/**
	 * C#-Parser class. The guts of this class are automatically generated
	 * by jay from the grammar rules and semantics below.
	 */
	class CSharpParser {
	    /// Currently constructed and parsed statement block.
	    Block currBlock;

	    /// Type of the most recent member/variable declaration.
	    Type currDeclType;

	    /// Particular stub class if currDeclType == stubType.
	    Class stubDeclClass;

	    /// Size of the recently declared array if currDeclType ==
	    /// arrayType.
	    int arrayDeclDim;

	    /// Element type of the recently declared array if currDeclType ==
	    /// arrayType.
	    Type arrayElementType;

	    /// Currently constructed and parsed method.
	    Method currMethod;

	    /// Currently constructed and parsed class.
	    Class currClass;

	    /// Currently constructed attribute in attribute list.
	    Attribute currAttr;

	   /** 
	    * Attributes of the most recent declaration. This is required
	    * since there can be more than one field in one declaration and we
	    * cannot pass the attributes directly as arguments "down" the
	    * rules.
            */
	    List attributes;

	    /** 
	     * Current loop depth. Used to determine the validity of a
	     * break/continue statement. 
	     */
	    int loopDepth;

	    /// Dictionary (string -> Class) of all known classes.
	    Hashtable classes;

	    /// Stub type constant for easier comparisons.
	    static Type stubType = ( new Class("", null, "" ) ).GetType();

	    /// Array type constant for easier comparisons.
	    static Type arrayType = ( new List() ).GetType();

	    /// Relevant modifiers
	    public enum Modifiers { PUBLIC, STATIC };

	    /// Modifiers of the most recent declaration.
	    IList modifiers;

	    /**
	     * Initialize the parser.
	     * @param classes Dictionary (string -> Class) of already parsed
	     *        classes; parsed classes are added to this.
	     */
	    public CSharpParser( Hashtable classes ) { 
		currBlock = null; 
		currClass = null;
		loopDepth = 0; 
		this.classes = classes;
	    }
	 
#line default

  /** simplified error message.
      @see <a href="#yyerror(java.lang.String, java.lang.String[])">yyerror</a>
    */
  public void yyerror (string message) {
    yyerror(message, null);
  }

  /** (syntax) error message.
      Can be overwritten to control message format.
      @param message text to be displayed.
      @param expected vector of acceptable tokens, if available.
    */
  public void yyerror (string message, string[] expected) {
    if ((expected != null) && (expected.Length  > 0)) {
      System.Console.Write (message+", expecting");
      for (int n = 0; n < expected.Length; ++ n)
        System.Console.Write (" "+expected[n]);
        System.Console.WriteLine ();
    } else
      System.Console.WriteLine (message);
  }

  /** debugging support, requires the package jay.yydebug.
      Set to null to suppress debugging messages.
    */
//t  protected yydebug.yyDebug debug;

  protected static  int yyFinal = 4;
//t  public static  string [] yyRule = {
//t    "$accept : compilation_unit",
//t    "compilation_unit : outer_declarations opt_EOF",
//t    "compilation_unit : opt_EOF",
//t    "opt_EOF :",
//t    "opt_EOF : EOF",
//t    "outer_declarations : outer_declaration",
//t    "outer_declarations : outer_declarations outer_declaration",
//t    "outer_declaration : using_directive",
//t    "outer_declaration : namespace_member_declaration",
//t    "using_directives : using_directive",
//t    "using_directives : using_directives using_directive",
//t    "using_directive : using_alias_directive",
//t    "using_directive : using_namespace_directive",
//t    "using_alias_directive : USING IDENTIFIER ASSIGN qualified_identifier SEMICOLON",
//t    "using_namespace_directive : USING qualified_identifier SEMICOLON",
//t    "namespace_member_declaration : type_declaration",
//t    "namespace_member_declaration : namespace_declaration",
//t    "namespace_declaration : opt_attributes NAMESPACE qualified_identifier namespace_body opt_semicolon",
//t    "namespace_body : OPEN_BRACE opt_using_directives opt_namespace_member_declarations CLOSE_BRACE",
//t    "opt_using_directives :",
//t    "opt_using_directives : using_directives",
//t    "opt_namespace_member_declarations :",
//t    "opt_namespace_member_declarations : namespace_member_declarations",
//t    "namespace_member_declarations : namespace_member_declaration",
//t    "namespace_member_declarations : namespace_member_declarations namespace_member_declaration",
//t    "qualified_identifier : IDENTIFIER",
//t    "qualified_identifier : qualified_identifier DOT IDENTIFIER",
//t    "opt_semicolon :",
//t    "opt_semicolon : SEMICOLON",
//t    "opt_comma :",
//t    "opt_comma : COMMA",
//t    "opt_identifier :",
//t    "opt_identifier : IDENTIFIER",
//t    "type : qualified_identifier",
//t    "type : builtin_types",
//t    "type : array_type",
//t    "builtin_types : OBJECT",
//t    "builtin_types : STRING",
//t    "builtin_types : BOOL",
//t    "builtin_types : DECIMAL",
//t    "builtin_types : FLOAT",
//t    "builtin_types : DOUBLE",
//t    "builtin_types : VOID",
//t    "builtin_types : CHAR",
//t    "builtin_types : integral_type",
//t    "integral_type : SBYTE",
//t    "integral_type : BYTE",
//t    "integral_type : SHORT",
//t    "integral_type : USHORT",
//t    "integral_type : INT",
//t    "integral_type : UINT",
//t    "integral_type : LONG",
//t    "integral_type : ULONG",
//t    "array_type : type rank_specifiers",
//t    "opt_rank_specifier :",
//t    "opt_rank_specifier : rank_specifiers",
//t    "rank_specifiers : rank_specifier",
//t    "rank_specifier : OPEN_BRACKET opt_dim_separators CLOSE_BRACKET",
//t    "opt_dim_separators :",
//t    "opt_dim_separators : dim_separators",
//t    "dim_separators : COMMA",
//t    "dim_separators : dim_separators COMMA",
//t    "literal : TRUE",
//t    "literal : FALSE",
//t    "literal : LITERAL_INTEGER",
//t    "literal : LITERAL_FLOAT",
//t    "literal : LITERAL_DOUBLE",
//t    "literal : LITERAL_DECIMAL",
//t    "literal : LITERAL_CHARACTER",
//t    "literal : LITERAL_STRING",
//t    "literal : NULL",
//t    "primary_expression : literal",
//t    "primary_expression : qualified_identifier",
//t    "primary_expression : this_access DOT qualified_identifier",
//t    "primary_expression : OPEN_PARENS expression CLOSE_PARENS",
//t    "primary_expression : invocation_expression",
//t    "primary_expression : this_access",
//t    "primary_expression : element_access",
//t    "primary_expression : post_increment_expression",
//t    "primary_expression : post_decrement_expression",
//t    "primary_expression : new_expression",
//t    "unary_expression : primary_expression",
//t    "unary_expression : BANG prefixed_unary_expression",
//t    "unary_expression : TILDE prefixed_unary_expression",
//t    "unary_expression : cast_expression",
//t    "cast_expression : OPEN_PARENS non_expression_type CLOSE_PARENS prefixed_unary_expression",
//t    "non_expression_type : builtin_types",
//t    "non_expression_type : non_expression_type rank_specifier",
//t    "non_expression_type : non_expression_type STAR",
//t    "prefixed_unary_expression : unary_expression",
//t    "prefixed_unary_expression : PLUS prefixed_unary_expression",
//t    "prefixed_unary_expression : MINUS prefixed_unary_expression",
//t    "prefixed_unary_expression : pre_increment_expression",
//t    "prefixed_unary_expression : pre_decrement_expression",
//t    "prefixed_unary_expression : STAR prefixed_unary_expression",
//t    "prefixed_unary_expression : BITWISE_AND prefixed_unary_expression",
//t    "multiplicative_expression : prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression STAR prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression DIV prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression PERCENT prefixed_unary_expression",
//t    "additive_expression : multiplicative_expression",
//t    "additive_expression : additive_expression PLUS multiplicative_expression",
//t    "additive_expression : additive_expression MINUS multiplicative_expression",
//t    "shift_expression : additive_expression",
//t    "shift_expression : shift_expression OP_SHIFT_LEFT additive_expression",
//t    "shift_expression : shift_expression OP_SHIFT_RIGHT additive_expression",
//t    "relational_expression : shift_expression",
//t    "relational_expression : relational_expression OP_LT shift_expression",
//t    "relational_expression : relational_expression OP_GT shift_expression",
//t    "relational_expression : relational_expression OP_LE shift_expression",
//t    "relational_expression : relational_expression OP_GE shift_expression",
//t    "relational_expression : relational_expression IS type",
//t    "relational_expression : relational_expression AS type",
//t    "equality_expression : relational_expression",
//t    "equality_expression : equality_expression OP_EQ relational_expression",
//t    "equality_expression : equality_expression OP_NE relational_expression",
//t    "and_expression : equality_expression",
//t    "and_expression : and_expression BITWISE_AND equality_expression",
//t    "exclusive_or_expression : and_expression",
//t    "exclusive_or_expression : exclusive_or_expression CARRET and_expression",
//t    "inclusive_or_expression : exclusive_or_expression",
//t    "inclusive_or_expression : inclusive_or_expression BITWISE_OR exclusive_or_expression",
//t    "conditional_and_expression : inclusive_or_expression",
//t    "conditional_and_expression : conditional_and_expression OP_AND inclusive_or_expression",
//t    "conditional_or_expression : conditional_and_expression",
//t    "conditional_or_expression : conditional_or_expression OP_OR conditional_and_expression",
//t    "conditional_expression : conditional_or_expression",
//t    "conditional_expression : conditional_or_expression INTERR expression COLON expression",
//t    "assignment_expression : primary_expression ASSIGN expression",
//t    "assignment_expression : primary_expression OP_MULT_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_DIV_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_ADD_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_SUB_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_MOD_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_SHIFT_LEFT_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_SHIFT_RIGHT_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_AND_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_OR_ASSIGN expression",
//t    "assignment_expression : primary_expression OP_XOR_ASSIGN expression",
//t    "expression : conditional_expression",
//t    "expression : assignment_expression",
//t    "post_increment_expression : primary_expression OP_INC",
//t    "post_decrement_expression : primary_expression OP_DEC",
//t    "pre_increment_expression : OP_INC prefixed_unary_expression",
//t    "pre_decrement_expression : OP_DEC prefixed_unary_expression",
//t    "invocation_expression : qualified_identifier OPEN_PARENS opt_argument_list CLOSE_PARENS",
//t    "opt_argument_list :",
//t    "opt_argument_list : argument_list",
//t    "argument_list : argument",
//t    "argument_list : argument_list COMMA argument",
//t    "argument : expression",
//t    "argument : REF expression",
//t    "argument : OUT expression",
//t    "new_expression : object_or_delegate_creation_expression",
//t    "new_expression : array_creation_expression",
//t    "$$1 :",
//t    "object_or_delegate_creation_expression : NEW type $$1 OPEN_PARENS opt_argument_list CLOSE_PARENS",
//t    "$$2 :",
//t    "array_creation_expression : NEW type OPEN_BRACKET int_expression_list CLOSE_BRACKET opt_rank_specifier $$2 opt_array_initializer",
//t    "array_creation_expression : NEW type array_initializer",
//t    "array_creation_expression : NEW type error",
//t    "opt_array_initializer :",
//t    "opt_array_initializer : array_initializer",
//t    "array_initializer : OPEN_BRACE CLOSE_BRACE",
//t    "array_initializer : OPEN_BRACE array_initializer_list opt_comma CLOSE_BRACE",
//t    "array_initializer_list : variable_initializer",
//t    "array_initializer_list : array_initializer_list COMMA variable_initializer",
//t    "constant_expression : expression",
//t    "int_expression_list : expression",
//t    "int_expression_list : int_expression_list COMMA expression",
//t    "element_access : primary_expression OPEN_BRACKET int_expression_list CLOSE_BRACKET",
//t    "this_access : THIS",
//t    "$$3 :",
//t    "block : OPEN_BRACE $$3 opt_statement_list CLOSE_BRACE",
//t    "opt_statement_list :",
//t    "opt_statement_list : statement_list",
//t    "statement_list : statement",
//t    "statement_list : statement_list statement",
//t    "statement : declaration_statement",
//t    "statement : embedded_statement",
//t    "statement : labeled_statement",
//t    "declaration_statement : local_variable_declaration SEMICOLON",
//t    "embedded_statement : block",
//t    "embedded_statement : empty_statement",
//t    "embedded_statement : expression_statement",
//t    "embedded_statement : selection_statement",
//t    "embedded_statement : iteration_statement",
//t    "embedded_statement : jump_statement",
//t    "embedded_statement : try_statement",
//t    "embedded_statement : checked_statement",
//t    "embedded_statement : unchecked_statement",
//t    "embedded_statement : lock_statement",
//t    "embedded_statement : unsafe_statement",
//t    "empty_statement : SEMICOLON",
//t    "labeled_statement : IDENTIFIER COLON statement",
//t    "expression_statement : statement_expression SEMICOLON",
//t    "statement_expression : invocation_expression",
//t    "statement_expression : object_or_delegate_creation_expression",
//t    "statement_expression : assignment_expression",
//t    "statement_expression : post_increment_expression",
//t    "statement_expression : post_decrement_expression",
//t    "statement_expression : pre_increment_expression",
//t    "statement_expression : pre_decrement_expression",
//t    "selection_statement : if_statement",
//t    "if_statement : IF OPEN_PARENS expression CLOSE_PARENS embedded_statement",
//t    "if_statement : IF OPEN_PARENS expression CLOSE_PARENS embedded_statement ELSE embedded_statement",
//t    "iteration_statement : while_statement",
//t    "iteration_statement : do_statement",
//t    "iteration_statement : for_statement",
//t    "$$4 :",
//t    "while_statement : WHILE $$4 OPEN_PARENS expression CLOSE_PARENS embedded_statement",
//t    "$$5 :",
//t    "do_statement : DO $$5 embedded_statement WHILE OPEN_PARENS expression CLOSE_PARENS SEMICOLON",
//t    "$$6 :",
//t    "for_statement : FOR OPEN_PARENS $$6 opt_for_initializer SEMICOLON opt_for_condition SEMICOLON opt_for_iterator CLOSE_PARENS embedded_statement",
//t    "opt_for_initializer :",
//t    "opt_for_initializer : for_initializer",
//t    "for_initializer : local_variable_declaration",
//t    "for_initializer : statement_expression_list",
//t    "opt_for_condition :",
//t    "opt_for_condition : expression",
//t    "opt_for_iterator :",
//t    "opt_for_iterator : statement_expression_list",
//t    "statement_expression_list : statement_expression",
//t    "statement_expression_list : statement_expression_list COMMA statement_expression",
//t    "jump_statement : break_statement",
//t    "jump_statement : continue_statement",
//t    "jump_statement : goto_statement",
//t    "jump_statement : return_statement",
//t    "jump_statement : throw_statement",
//t    "break_statement : BREAK SEMICOLON",
//t    "continue_statement : CONTINUE SEMICOLON",
//t    "return_statement : RETURN SEMICOLON",
//t    "return_statement : RETURN expression SEMICOLON",
//t    "lock_statement : LOCK OPEN_PARENS expression CLOSE_PARENS embedded_statement",
//t    "goto_statement : GOTO IDENTIFIER SEMICOLON",
//t    "goto_statement : GOTO CASE constant_expression SEMICOLON",
//t    "goto_statement : GOTO DEFAULT SEMICOLON",
//t    "throw_statement : THROW expression SEMICOLON",
//t    "throw_statement : THROW SEMICOLON",
//t    "try_statement : TRY block catch_clauses",
//t    "try_statement : TRY block opt_catch_clauses FINALLY block",
//t    "try_statement : TRY block error",
//t    "opt_catch_clauses :",
//t    "opt_catch_clauses : catch_clauses",
//t    "catch_clauses : catch_clause",
//t    "catch_clauses : catch_clauses catch_clause",
//t    "catch_clause : CATCH opt_catch_args block",
//t    "opt_catch_args :",
//t    "opt_catch_args : catch_args",
//t    "catch_args : OPEN_PARENS type opt_identifier CLOSE_PARENS",
//t    "checked_statement : CHECKED block",
//t    "unchecked_statement : UNCHECKED block",
//t    "unsafe_statement : UNSAFE block",
//t    "local_variable_type : builtin_types opt_rank_specifier",
//t    "local_variable_declaration : local_variable_type variable_declarators",
//t    "variable_declarators : variable_declarator",
//t    "variable_declarators : variable_declarators COMMA variable_declarator",
//t    "variable_declarator : IDENTIFIER ASSIGN variable_initializer",
//t    "variable_declarator : IDENTIFIER",
//t    "variable_initializer : expression",
//t    "variable_initializer : array_initializer",
//t    "method_declaration : method_header block",
//t    "method_declaration : method_header SEMICOLON",
//t    "method_header : opt_attributes opt_modifiers type IDENTIFIER OPEN_PARENS opt_formal_parameter_list CLOSE_PARENS",
//t    "opt_formal_parameter_list :",
//t    "opt_formal_parameter_list : formal_parameter_list",
//t    "formal_parameter_list : fixed_parameters",
//t    "fixed_parameters : fixed_parameter",
//t    "fixed_parameters : fixed_parameters COMMA fixed_parameter",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier type IDENTIFIER",
//t    "opt_parameter_modifier :",
//t    "opt_parameter_modifier : parameter_modifier",
//t    "parameter_modifier : REF",
//t    "parameter_modifier : OUT",
//t    "type_declaration : class_declaration",
//t    "$$7 :",
//t    "class_declaration : opt_attributes opt_modifiers CLASS IDENTIFIER opt_class_base $$7 class_body opt_semicolon",
//t    "opt_class_base :",
//t    "opt_class_base : class_base",
//t    "class_base : COLON qualified_identifier",
//t    "class_body : OPEN_BRACE opt_class_member_declarations CLOSE_BRACE",
//t    "opt_class_member_declarations :",
//t    "opt_class_member_declarations : class_member_declarations",
//t    "class_member_declarations : class_member_declaration",
//t    "class_member_declarations : class_member_declarations class_member_declaration",
//t    "class_member_declaration : method_declaration",
//t    "class_member_declaration : field_declaration",
//t    "class_member_declaration : constructor_declaration",
//t    "field_declaration : opt_attributes opt_modifiers type variable_declarators SEMICOLON",
//t    "opt_modifiers :",
//t    "opt_modifiers : modifiers",
//t    "modifiers : modifier",
//t    "modifiers : modifiers modifier",
//t    "modifier : NEW",
//t    "modifier : PUBLIC",
//t    "modifier : PROTECTED",
//t    "modifier : INTERNAL",
//t    "modifier : PRIVATE",
//t    "modifier : ABSTRACT",
//t    "modifier : SEALED",
//t    "modifier : STATIC",
//t    "modifier : READONLY",
//t    "modifier : VIRTUAL",
//t    "modifier : OVERRIDE",
//t    "modifier : EXTERN",
//t    "modifier : VOLATILE",
//t    "modifier : UNSAFE",
//t    "constructor_declaration : constructor_header block",
//t    "constructor_declaration : constructor_header SEMICOLON",
//t    "constructor_header : opt_attributes opt_modifiers IDENTIFIER OPEN_PARENS opt_formal_parameter_list CLOSE_PARENS",
//t    "opt_attributes :",
//t    "opt_attributes : attribute_sections",
//t    "attribute_sections : attribute_section",
//t    "attribute_sections : attribute_sections attribute_section",
//t    "attribute_section : OPEN_BRACKET attribute_list opt_comma CLOSE_BRACKET",
//t    "attribute_list : attribute",
//t    "attribute_list : attribute_list COMMA attribute",
//t    "$$8 :",
//t    "attribute : qualified_identifier $$8 opt_attribute_arguments",
//t    "opt_attribute_arguments :",
//t    "opt_attribute_arguments : OPEN_PARENS attribute_arguments CLOSE_PARENS",
//t    "attribute_arguments : opt_positional_argument_list",
//t    "attribute_arguments : positional_argument_list COMMA named_argument_list",
//t    "attribute_arguments : named_argument_list",
//t    "opt_positional_argument_list :",
//t    "opt_positional_argument_list : positional_argument_list",
//t    "positional_argument_list : positional_argument",
//t    "positional_argument_list : positional_argument_list COMMA positional_argument",
//t    "positional_argument : expression",
//t    "named_argument_list : named_argument",
//t    "named_argument_list : named_argument_list COMMA named_argument",
//t    "named_argument : IDENTIFIER ASSIGN expression",
//t  };
  protected static  string [] yyNames = {    
    "end-of-file",null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,"'!'",null,null,null,"'%'","'&'",
    null,"'('","')'","'*'","'+'","','","'-'","'.'","'/'",null,null,null,
    null,null,null,null,null,null,null,"':'","';'","'<'","'='","'>'",
    "'?'",null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,"'['",null,"']'","'^'",null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,"'{'","'|'","'}'","'~'",null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,"EOF","NONE","ERROR","FIRST_KEYWORD","ABSTRACT","AS","ADD",
    "ASSEMBLY","BASE","BOOL","BREAK","BYTE","CASE","CATCH","CHAR",
    "CHECKED","CLASS","CONST","CONTINUE","DECIMAL","DEFAULT","DELEGATE",
    "DO","DOUBLE","ELSE","ENUM","EVENT","EXPLICIT","EXTERN","FALSE",
    "FINALLY","FIXED","FLOAT","FOR","FOREACH","GOTO","IF","IMPLICIT","IN",
    "INT","INTERFACE","INTERNAL","IS","LOCK","LONG","NAMESPACE","NEW",
    "NULL","OBJECT","OPERATOR","OUT","OVERRIDE","PARAMS","PRIVATE",
    "PROTECTED","PUBLIC","READONLY","REF","RETURN","REMOVE","SBYTE",
    "SEALED","SHORT","SIZEOF","STACKALLOC","STATIC","STRING","STRUCT",
    "SWITCH","THIS","THROW","TRUE","TRY","TYPEOF","UINT","ULONG",
    "UNCHECKED","UNSAFE","USHORT","USING","VIRTUAL","VOID","VOLATILE",
    "WHILE","YIELD","GET","\"get\"","SET","\"set\"","LAST_KEYWORD",
    "OPEN_BRACE","CLOSE_BRACE","OPEN_BRACKET","CLOSE_BRACKET",
    "OPEN_PARENS","CLOSE_PARENS","DOT","COMMA","COLON","SEMICOLON",
    "TILDE","PLUS","MINUS","BANG","ASSIGN","OP_LT","OP_GT","BITWISE_AND",
    "BITWISE_OR","STAR","PERCENT","DIV","CARRET","INTERR","OP_INC",
    "\"++\"","OP_DEC","\"--\"","OP_SHIFT_LEFT","\"<<\"","OP_SHIFT_RIGHT",
    "\">>\"","OP_LE","\"<=\"","OP_GE","\">=\"","OP_EQ","\"==\"","OP_NE",
    "\"!=\"","OP_AND","\"&&\"","OP_OR","\"||\"","OP_MULT_ASSIGN","\"*=\"",
    "OP_DIV_ASSIGN","\"/=\"","OP_MOD_ASSIGN","\"%=\"","OP_ADD_ASSIGN",
    "\"+=\"","OP_SUB_ASSIGN","\"-=\"","OP_SHIFT_LEFT_ASSIGN","\"<<=\"",
    "OP_SHIFT_RIGHT_ASSIGN","\">>=\"","OP_AND_ASSIGN","\"&=\"",
    "OP_XOR_ASSIGN","\"^=\"","OP_OR_ASSIGN","\"|=\"","OP_PTR","\"->\"",
    "LITERAL_INTEGER","\"int literal\"","LITERAL_FLOAT",
    "\"float literal\"","LITERAL_DOUBLE","\"double literal\"",
    "LITERAL_DECIMAL","\"decimal literal\"","LITERAL_CHARACTER",
    "\"character literal\"","LITERAL_STRING","\"string literal\"",
    "IDENTIFIER","CLOSE_PARENS_CAST","CLOSE_PARENS_NO_CAST",
    "CLOSE_PARENS_OPEN_PARENS","CLOSE_PARENS_MINUS","LOWPREC","UMINUS",
    "HIGHPREC",
  };

  /** index-checked interface to yyNames[].
      @param token single character or %token value.
      @return token name or [illegal] or [unknown].
    */
//t  public static string yyname (int token) {
//t    if ((token < 0) || (token > yyNames.Length)) return "[illegal]";
//t    string name;
//t    if ((name = yyNames[token]) != null) return name;
//t    return "[unknown]";
//t  }

  /** computes list of expected tokens on error by tracing the tables.
      @param state for which to compute the list.
      @return list of token names.
    */
  protected string[] yyExpecting (int state) {
    int token, n, len = 0;
    bool[] ok = new bool[yyNames.Length];

    if ((n = yySindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }
    if ((n = yyRindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }

    string [] result = new string[len];
    for (n = token = 0; n < len;  ++ token)
      if (ok[token]) result[n++] = yyNames[token];
    return result;
  }

  /** the generated parser, with debugging messages.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @param yydebug debug message writer implementing yyDebug, or null.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  public Object yyparse (yyParser.yyInput yyLex, Object yyd)
				 {
//t    this.debug = (yydebug.yyDebug)yyd;
    return yyparse(yyLex);
  }

  /** initial size and increment of the state/value stack [default 256].
      This is not final so that it can be overwritten outside of invocations
      of yyparse().
    */
  protected int yyMax;

  /** executed at the beginning of a reduce action.
      Used as $$ = yyDefault($1), prior to the user-specified action, if any.
      Can be overwritten to provide deep copy, etc.
      @param first value for $1, or null.
      @return first.
    */
  protected Object yyDefault (Object first) {
    return first;
  }

  /** the generated parser.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  public Object yyparse (yyParser.yyInput yyLex)
				{
    if (yyMax <= 0) yyMax = 256;			// initial size
    int yyState = 0;                                   // state stack ptr
    int [] yyStates = new int[yyMax];	                // state stack 
    Object yyVal = null;                               // value stack ptr
    Object [] yyVals = new Object[yyMax];	        // value stack
    int yyToken = -1;					// current input
    int yyErrorFlag = 0;				// #tks to shift

    int yyTop = 0;
    goto skip;
    yyLoop:
    yyTop++;
    skip:
    for (;; ++ yyTop) {
      if (yyTop >= yyStates.Length) {			// dynamically increase
        int[] i = new int[yyStates.Length+yyMax];
        yyStates.CopyTo (i, 0);
        yyStates = i;
        Object[] o = new Object[yyVals.Length+yyMax];
        yyVals.CopyTo (o, 0);
        yyVals = o;
      }
      yyStates[yyTop] = yyState;
      yyVals[yyTop] = yyVal;
//t      if (debug != null) debug.push(yyState, yyVal);

      yyDiscarded: for (;;) {	// discarding a token does not change stack
        int yyN;
        if ((yyN = yyDefRed[yyState]) == 0) {	// else [default] reduce (yyN)
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t              debug.lex(yyState, yyToken, yyname(yyToken), yyLex.value());
          }
          if ((yyN = yySindex[yyState]) != 0 && ((yyN += yyToken) >= 0)
              && (yyN < yyTable.Length) && (yyCheck[yyN] == yyToken)) {
//t            if (debug != null)
//t              debug.shift(yyState, yyTable[yyN], yyErrorFlag-1);
            yyState = yyTable[yyN];		// shift to yyN
            yyVal = yyLex.value();
            yyToken = -1;
            if (yyErrorFlag > 0) -- yyErrorFlag;
            goto yyLoop;
          }
          if ((yyN = yyRindex[yyState]) != 0 && (yyN += yyToken) >= 0
              && yyN < yyTable.Length && yyCheck[yyN] == yyToken)
            yyN = yyTable[yyN];			// reduce (yyN)
          else
            switch (yyErrorFlag) {
  
            case 0:
              yyerror("syntax error", yyExpecting(yyState));
//t              if (debug != null) debug.error("syntax error");
              goto case 1;
            case 1: case 2:
              yyErrorFlag = 3;
              do {
                if ((yyN = yySindex[yyStates[yyTop]]) != 0
                    && (yyN += Token.yyErrorCode) >= 0 && yyN < yyTable.Length
                    && yyCheck[yyN] == Token.yyErrorCode) {
//t                  if (debug != null)
//t                    debug.shift(yyStates[yyTop], yyTable[yyN], 3);
                  yyState = yyTable[yyN];
                  yyVal = yyLex.value();
                  goto yyLoop;
                }
//t                if (debug != null) debug.pop(yyStates[yyTop]);
              } while (-- yyTop >= 0);
//t              if (debug != null) debug.reject();
              throw new yyParser.yyException("irrecoverable syntax error");
  
            case 3:
              if (yyToken == 0) {
//t                if (debug != null) debug.reject();
                throw new yyParser.yyException("irrecoverable syntax error at end-of-file");
              }
//t              if (debug != null)
//t                debug.discard(yyState, yyToken, yyname(yyToken),
//t  							yyLex.value());
              yyToken = -1;
              goto yyDiscarded;		// leave stack alone
            }
        }
        int yyV = yyTop + 1-yyLen[yyN];
//t        if (debug != null)
//t          debug.reduce(yyState, yyStates[yyV-1], yyN, yyRule[yyN], yyLen[yyN]);
        yyVal = yyDefault(yyV > yyTop ? null : yyVals[yyV]);
        switch (yyN) {
case 1:
#line 297 "cs-parser.jay"
  { yyVal = classes; }
  break;
case 2:
#line 298 "cs-parser.jay"
  { yyVal = classes; }
  break;
case 25:
#line 369 "cs-parser.jay"
  { yyVal = new QID( (string)yyVals[0+yyTop] ); }
  break;
case 26:
#line 370 "cs-parser.jay"
  { ( (QID) yyVals[-2+yyTop] ).add( (string) yyVals[0+yyTop]); yyVal = yyVals[-2+yyTop]; }
  break;
case 31:
#line 384 "cs-parser.jay"
  { yyVal = null; }
  break;
case 33:
#line 396 "cs-parser.jay"
  { 
	    stubDeclClass = (Class) classes[yyVals[0+yyTop].ToString()];
	    if( stubDeclClass == null )
		errUnknownType();
	    yyVal = currDeclType = stubType;
	  }
  break;
case 34:
#line 402 "cs-parser.jay"
  { stubDeclClass = null; }
  break;
case 36:
#line 415 "cs-parser.jay"
  { yyVal = currDeclType = (new Object()).GetType(); }
  break;
case 37:
#line 416 "cs-parser.jay"
  { yyVal = currDeclType = "".GetType(); }
  break;
case 38:
#line 417 "cs-parser.jay"
  { yyVal = currDeclType = false.GetType(); }
  break;
case 39:
#line 418 "cs-parser.jay"
  { yyVal = currDeclType = 0.0d.GetType(); }
  break;
case 40:
#line 419 "cs-parser.jay"
  { yyVal = currDeclType = 0.0d.GetType(); }
  break;
case 41:
#line 420 "cs-parser.jay"
  { yyVal = currDeclType = 0.0d.GetType(); }
  break;
case 42:
#line 421 "cs-parser.jay"
  { yyVal = null; }
  break;
case 43:
#line 422 "cs-parser.jay"
  { yyVal = 'a'.GetType(); }
  break;
case 44:
#line 423 "cs-parser.jay"
  { yyVal = currDeclType = 0.GetType(); }
  break;
case 53:
#line 439 "cs-parser.jay"
  { yyVal = currDeclType = arrayType; arrayDeclDim = (int) yyVals[0+yyTop]; arrayElementType = (Type) yyVals[-1+yyTop]; }
  break;
case 54:
#line 445 "cs-parser.jay"
  { yyVal = 0; }
  break;
case 57:
#line 458 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 58:
#line 462 "cs-parser.jay"
  { yyVal = 1; }
  break;
case 60:
#line 467 "cs-parser.jay"
  { yyVal = 2; }
  break;
case 61:
#line 468 "cs-parser.jay"
  { yyVal = (int) yyVals[-1+yyTop] + 1; }
  break;
case 62:
#line 480 "cs-parser.jay"
  { yyVal = new BoolExpr( true ); }
  break;
case 63:
#line 481 "cs-parser.jay"
  { yyVal = new BoolExpr( false ); }
  break;
case 64:
#line 482 "cs-parser.jay"
  { yyVal = new IntExpr( (int) yyVals[0+yyTop] ); }
  break;
case 65:
#line 483 "cs-parser.jay"
  { yyVal = new FloatExpr( (float) yyVals[0+yyTop] ); }
  break;
case 66:
#line 484 "cs-parser.jay"
  { yyVal = new FloatExpr( (double) yyVals[0+yyTop] ); }
  break;
case 67:
#line 485 "cs-parser.jay"
  { yyVal = new FloatExpr( (double) yyVals[0+yyTop] ); }
  break;
case 68:
#line 486 "cs-parser.jay"
  { yyVal = new StringExpr( yyVals[0+yyTop].ToString() ); }
  break;
case 69:
#line 487 "cs-parser.jay"
  { yyVal = new StringExpr( (string) yyVals[0+yyTop] ); }
  break;
case 70:
#line 488 "cs-parser.jay"
  { yyVal = new NullExpr(); }
  break;
case 72:
#line 494 "cs-parser.jay"
  {
	    Ref r = new Ref( (QID) yyVals[0+yyTop], false );
	    VariableDecl v = resolveVar( r );
	    if( v != null )
		yyVal = makeRefExpr( v, r );
	    else
		yyVal = null; 
	     /* if v == null, resolveVar() has already reported error*/
	  }
  break;
case 73:
#line 506 "cs-parser.jay"
  {
	    Ref r = new Ref( (QID) yyVals[0+yyTop], true );
	    VariableDecl v = resolveVar( r );
	    if( v != null )
		yyVal = makeRefExpr( v, r );
	    else
		yyVal = null; 
	    /* if v == null, resolveVar() has already reported error*/
	  }
  break;
case 74:
#line 516 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 82:
#line 535 "cs-parser.jay"
  { chkb(yyVals[0+yyTop]); yyVal = new BoolExpr( BoolExpr.Op.NOT, (BoolExpr) yyVals[0+yyTop] ); }
  break;
case 83:
#line 537 "cs-parser.jay"
  { chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.TILDE, (IntExpr) yyVals[0+yyTop] ); }
  break;
case 85:
#line 551 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; }
  break;
case 90:
#line 571 "cs-parser.jay"
  { if( yyVals[0+yyTop] is IntExpr || yyVals[0+yyTop] is FloatExpr ) yyVal = yyVals[0+yyTop]; else errNumExp(); }
  break;
case 91:
#line 573 "cs-parser.jay"
  { 
	    if( yyVals[0+yyTop] is IntExpr )
		yyVal = new IntExpr( IntExpr.Op.NEG, (IntExpr) yyVals[0+yyTop] );
	    else if( yyVals[0+yyTop] is FloatExpr )
		yyVal = new FloatExpr( FloatExpr.Op.NEG, (FloatExpr) yyVals[0+yyTop] );
	    else
		errNumExp();
	  }
  break;
case 94:
#line 584 "cs-parser.jay"
  { errUnimplOp(); }
  break;
case 95:
#line 586 "cs-parser.jay"
  { errUnimplOp(); }
  break;
case 97:
#line 592 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr ) { 
		chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.MUL, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); 
	    } else if( yyVals[-2+yyTop] is FloatExpr ) {
		chkf(yyVals[0+yyTop]); yyVal = new FloatExpr( FloatExpr.Op.MUL, (FloatExpr) yyVals[-2+yyTop], (FloatExpr) yyVals[0+yyTop] ); 
	    } else
		errNumExp();
	  }
  break;
case 98:
#line 601 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr ) { 
		chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.DIV, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); 
	    } else if( yyVals[-2+yyTop] is FloatExpr ) {
		chkf(yyVals[0+yyTop]); yyVal = new FloatExpr( FloatExpr.Op.DIV, (FloatExpr) yyVals[-2+yyTop], (FloatExpr) yyVals[0+yyTop] ); 
	    } else
		errNumExp();
	  }
  break;
case 99:
#line 610 "cs-parser.jay"
  { chki(yyVals[-2+yyTop]); chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.MOD, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); }
  break;
case 101:
#line 616 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr ) { 
		chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.ADD, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); 
	    } else if( yyVals[-2+yyTop] is FloatExpr ) {
		chkf(yyVals[0+yyTop]); yyVal = new FloatExpr( FloatExpr.Op.ADD, (FloatExpr) yyVals[-2+yyTop], (FloatExpr) yyVals[0+yyTop] ); 
	    } else if( yyVals[-2+yyTop] is StringExpr ) {
		chks(yyVals[0+yyTop]); yyVal = new StringExpr( StringExpr.Op.CONCAT, (StringExpr) yyVals[-2+yyTop], (StringExpr) yyVals[0+yyTop] );
	    } else
		errNumExp();
	  }
  break;
case 102:
#line 627 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr ) { 
		chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.SUB, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); 
	    } else if( yyVals[-2+yyTop] is FloatExpr ) {
		chkf(yyVals[0+yyTop]); yyVal = new FloatExpr( FloatExpr.Op.SUB, (FloatExpr) yyVals[-2+yyTop], (FloatExpr) yyVals[0+yyTop] ); 
	    } else
		errNumExp();
	  }
  break;
case 104:
#line 640 "cs-parser.jay"
  { chki(yyVals[-2+yyTop]); chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.SHL, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); }
  break;
case 105:
#line 642 "cs-parser.jay"
  { chki(yyVals[-2+yyTop]); chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.SHR, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); }
  break;
case 107:
#line 648 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr || yyVals[-2+yyTop] is FloatExpr ) { 
		chksametype( yyVals[-2+yyTop], yyVals[0+yyTop]); yyVal = new BoolExpr( BoolExpr.Op.LT, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ); 
	    } else
		errNumExp();
	  }
  break;
case 108:
#line 655 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr || yyVals[-2+yyTop] is FloatExpr ) { 
		chksametype( yyVals[-2+yyTop], yyVals[0+yyTop]); yyVal = new BoolExpr( BoolExpr.Op.GT, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ); 
	    } else
		errNumExp();
	  }
  break;
case 109:
#line 662 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr || yyVals[-2+yyTop] is FloatExpr ) { 
		chksametype( yyVals[-2+yyTop], yyVals[0+yyTop]); yyVal = new BoolExpr( BoolExpr.Op.LE, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ); 
	    } else
		errNumExp();
	  }
  break;
case 110:
#line 669 "cs-parser.jay"
  { 
	    if( yyVals[-2+yyTop] is IntExpr || yyVals[-2+yyTop] is FloatExpr ) { 
		chksametype( yyVals[-2+yyTop], yyVals[0+yyTop]); yyVal = new BoolExpr( BoolExpr.Op.GE, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ); 
	    } else
		errNumExp();
	  }
  break;
case 111:
#line 676 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 112:
#line 678 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 114:
#line 684 "cs-parser.jay"
  { if( !( yyVals[-2+yyTop] is NullExpr || yyVals[0+yyTop] is NullExpr ) ) chksametype( yyVals[-2+yyTop], yyVals[0+yyTop] );
	    yyVal = new BoolExpr( BoolExpr.Op.EQ, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ); }
  break;
case 115:
#line 687 "cs-parser.jay"
  { if( !( yyVals[-2+yyTop] is NullExpr || yyVals[0+yyTop] is NullExpr ) ) chksametype( yyVals[-2+yyTop], yyVals[0+yyTop] );
            yyVal = new BoolExpr( BoolExpr.Op.NEQ, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ); }
  break;
case 117:
#line 694 "cs-parser.jay"
  { chki(yyVals[-2+yyTop]); chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.AND, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); }
  break;
case 119:
#line 700 "cs-parser.jay"
  { chki(yyVals[-2+yyTop]); chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.XOR, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); }
  break;
case 121:
#line 706 "cs-parser.jay"
  { chki(yyVals[-2+yyTop]); chki(yyVals[0+yyTop]); yyVal = new IntExpr( IntExpr.Op.OR, (IntExpr) yyVals[-2+yyTop], (IntExpr) yyVals[0+yyTop] ); }
  break;
case 123:
#line 712 "cs-parser.jay"
  { chkb(yyVals[-2+yyTop]); chkb(yyVals[0+yyTop]); yyVal = new BoolExpr( BoolExpr.Op.AND, (BoolExpr) yyVals[-2+yyTop], (BoolExpr) yyVals[0+yyTop] ); }
  break;
case 125:
#line 718 "cs-parser.jay"
  { chkb(yyVals[-2+yyTop]); chkb(yyVals[0+yyTop]); yyVal = new BoolExpr( BoolExpr.Op.OR, (BoolExpr) yyVals[-2+yyTop], (BoolExpr) yyVals[0+yyTop] ); }
  break;
case 127:
#line 724 "cs-parser.jay"
  { errUnimplOp(); }
  break;
case 128:
#line 729 "cs-parser.jay"
  { yyVal = makeAssign(  yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ); }
  break;
case 129:
#line 731 "cs-parser.jay"
  { if( yyVals[0+yyTop] is IntExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.MUL, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else if( yyVals[0+yyTop] is FloatExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new FloatExpr( FloatExpr.Op.MUL, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else
		errNumExp();
	  }
  break;
case 130:
#line 739 "cs-parser.jay"
  { if( yyVals[0+yyTop] is IntExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.DIV, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else if( yyVals[0+yyTop] is FloatExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new FloatExpr( FloatExpr.Op.DIV, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else
		errNumExp();
	  }
  break;
case 131:
#line 748 "cs-parser.jay"
  { if( yyVals[0+yyTop] is IntExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.ADD, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else if( yyVals[0+yyTop] is FloatExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new FloatExpr( FloatExpr.Op.ADD, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else if( yyVals[0+yyTop] is StringExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new StringExpr( StringExpr.Op.CONCAT, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else
		errNumExp();
	  }
  break;
case 132:
#line 759 "cs-parser.jay"
  { if( yyVals[0+yyTop] is IntExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.SUB, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else if( yyVals[0+yyTop] is FloatExpr )
		yyVal = makeAssign( yyVals[-2+yyTop], new FloatExpr( FloatExpr.Op.SUB, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) );
	    else
		errNumExp();
	  }
  break;
case 133:
#line 768 "cs-parser.jay"
  { chki( yyVals[0+yyTop]); yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.MOD, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) ); }
  break;
case 134:
#line 770 "cs-parser.jay"
  { chki( yyVals[0+yyTop]); yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.SHL, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) ); }
  break;
case 135:
#line 772 "cs-parser.jay"
  { chki( yyVals[0+yyTop]); yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.SHR, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) ); }
  break;
case 136:
#line 774 "cs-parser.jay"
  { chki( yyVals[0+yyTop]); yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.AND, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) ); }
  break;
case 137:
#line 776 "cs-parser.jay"
  { chki( yyVals[0+yyTop]); yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.OR, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) ); }
  break;
case 138:
#line 778 "cs-parser.jay"
  { chki( yyVals[0+yyTop]); yyVal = makeAssign( yyVals[-2+yyTop], new IntExpr( IntExpr.Op.XOR, (Expr) yyVals[-2+yyTop], (Expr) yyVals[0+yyTop] ) ); }
  break;
case 141:
#line 788 "cs-parser.jay"
  { chki(yyVals[-1+yyTop]); 
	    yyVal = new IntExpr( IntExpr.Op.SUB, 
		makeAssign( yyVals[-1+yyTop], new IntExpr( IntExpr.Op.ADD, (Expr) yyVals[-1+yyTop], new IntExpr( 1 ) ) ),
		new IntExpr( 1 ) );
	  }
  break;
case 142:
#line 797 "cs-parser.jay"
  { chki(yyVals[-1+yyTop]); 
	    yyVal = new IntExpr( IntExpr.Op.ADD, 
		makeAssign( yyVals[-1+yyTop], new IntExpr( IntExpr.Op.SUB, (Expr) yyVals[-1+yyTop], new IntExpr( 1 ) ) ),
		new IntExpr( 1 ) );
	  }
  break;
case 143:
#line 806 "cs-parser.jay"
  { chki(yyVals[0+yyTop]); 
	    yyVal = makeAssign( yyVals[0+yyTop], new IntExpr( IntExpr.Op.ADD, (Expr) yyVals[0+yyTop], new IntExpr( 1 ) ) );
	  }
  break;
case 144:
#line 813 "cs-parser.jay"
  { chki(yyVals[0+yyTop]); 
	    yyVal = makeAssign( yyVals[0+yyTop], new IntExpr( IntExpr.Op.SUB, (Expr) yyVals[0+yyTop], new IntExpr( 1 ) ) );
	  }
  break;
case 145:
#line 820 "cs-parser.jay"
  {
	    Ref r = new Ref( (QID) yyVals[-3+yyTop], false );
	    Method m = resolveMethod( r, (IList) yyVals[-1+yyTop] );
	    if( m != null )
		yyVal = makeCall( r, m, (List) yyVals[-1+yyTop] );
          }
  break;
case 146:
#line 840 "cs-parser.jay"
  { yyVal = new List(0); }
  break;
case 148:
#line 846 "cs-parser.jay"
  { IList list = new List(4); list.Add( yyVals[0+yyTop] ); yyVal = list; }
  break;
case 149:
#line 848 "cs-parser.jay"
  { IList list = (IList) yyVals[-2+yyTop]; list.Add (yyVals[0+yyTop]); yyVal = list; }
  break;
case 151:
#line 853 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 152:
#line 854 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 155:
#line 863 "cs-parser.jay"
  { if( currDeclType != stubType ) errNewType(); }
  break;
case 156:
#line 865 "cs-parser.jay"
  { 
	    if( stubDeclClass.constructible( (List) yyVals[-1+yyTop] ) )
		yyVal = new StubExpr( stubDeclClass, (List) yyVals[-1+yyTop] ); 
	    else {
		errNoMatchingCtor();
		yyVal = null;
	    }
	  }
  break;
case 157:
#line 877 "cs-parser.jay"
  { /* opt_array_initializer wants an array type, but we have a plain one here*/
	    arrayElementType = currDeclType; currDeclType = arrayType; }
  break;
case 158:
#line 880 "cs-parser.jay"
  { if( yyVals[0+yyTop] == null ) {
		Expr s = (Expr) ( (IList) yyVals[-4+yyTop] )[0];
		yyVal = new ArrayExpr( (Type) yyVals[-6+yyTop], (IntExpr) s );
		/* TODO: support multiple dimensions here*/
	    } else
		yyVal = yyVals[0+yyTop]; /* treat init'ed new like a literal*/
	  }
  break;
case 159:
#line 889 "cs-parser.jay"
  { if( (Type) yyVals[-1+yyTop] != arrayType ) {
		errNewSyntax();
		yyVal = null;
	    } else
		yyVal = yyVals[0+yyTop]; 
	  }
  break;
case 160:
#line 896 "cs-parser.jay"
  { errNewSyntax(); }
  break;
case 161:
#line 900 "cs-parser.jay"
  { yyVal = null; }
  break;
case 163:
#line 906 "cs-parser.jay"
  { yyVal = new ArrayExpr( new List(), arrayElementType ); }
  break;
case 164:
#line 908 "cs-parser.jay"
  { yyVal = new ArrayExpr( (List) yyVals[-2+yyTop], arrayElementType ); }
  break;
case 165:
#line 913 "cs-parser.jay"
  { 
	    IList list = new List( 4 ); 

	    /* don't check types of array-type elements*/
	    if( yyVals[0+yyTop] is IList || ((Expr) yyVals[0+yyTop]).type() == arrayElementType )
		list.Add ( ( (Expr) yyVals[0+yyTop] ).eval() ); 
	    else
		errAssignmentTypeMismatch();

	    yyVal = list; 
	  }
  break;
case 166:
#line 925 "cs-parser.jay"
  { 
	    /* don't check types of array-type elements*/
	    if( !( yyVals[0+yyTop] is IList ) && ( (Expr) yyVals[0+yyTop] ).type() != arrayElementType )
		errAssignmentTypeMismatch();
	    else
		( (IList) yyVals[-2+yyTop] ).Add ( ( (Expr) yyVals[0+yyTop] ).eval() );
	    yyVal = yyVals[-2+yyTop]; 
	  }
  break;
case 168:
#line 941 "cs-parser.jay"
  { chki( yyVals[0+yyTop] ); IList list = new List (4); list.Add (yyVals[0+yyTop]); yyVal = list; }
  break;
case 169:
#line 943 "cs-parser.jay"
  { chki( yyVals[0+yyTop] ); ( (IList) yyVals[-2+yyTop] ).Add (yyVals[0+yyTop]); yyVal = yyVals[-2+yyTop]; }
  break;
case 170:
#line 948 "cs-parser.jay"
  {
	    ArrayExpr ar = yyVals[-3+yyTop] as ArrayExpr;
	    List idx = (List) yyVals[-1+yyTop];
	    IntExpr ind = (IntExpr) ( idx[0] );
	    if( idx.Count != 1 ) {
		errOneDimArrayAccess();
		yyVal = null;
	    } else if( ar == null ) {
		errNonArrayElementAccess( yyVals[-3+yyTop].ToString() );
		yyVal = null;
	    } else {
		if( ar.elementType() == false.GetType() )
		    yyVal = new BoolExpr( ar, ind );
		else if( ar.elementType() == 0.GetType() )
		    yyVal = new IntExpr( ar, ind );
		else if( ar.elementType() == (0.0d).GetType() )
		    yyVal = new FloatExpr( ar, ind );
		else if( ar.elementType() == "".GetType() )
		    yyVal = new StringExpr( ar, ind );
		else if( ar.elementType() == stubType )
		    yyVal = new StubExpr( ar, ind );
		else if( ar.elementType() == arrayType )
		    yyVal = new ArrayExpr( ar, ind );
		else
		    throw new Exception( "CSharpParser.element_access: invalid expression type " + ar.elementType() );
	    }
          }
  break;
case 171:
#line 979 "cs-parser.jay"
  { yyVal = new ThisExpr( currClass ); }
  break;
case 172:
#line 999 "cs-parser.jay"
  { currBlock = new Block( currBlock ); }
  break;
case 173:
#line 1001 "cs-parser.jay"
  {  yyVal = currBlock; currBlock = currBlock.getParent(); }
  break;
case 179:
#line 1017 "cs-parser.jay"
  { if( yyVals[0+yyTop] != null ) currBlock.addStm( (Statement) yyVals[0+yyTop] ); }
  break;
case 193:
#line 1043 "cs-parser.jay"
  { yyVal = null; }
  break;
case 194:
#line 1048 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; }
  break;
case 195:
#line 1052 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 196:
#line 1060 "cs-parser.jay"
  { yyVal = new ExprStm( (Expr) yyVals[0+yyTop] ); }
  break;
case 197:
#line 1061 "cs-parser.jay"
  { yyVal = new ExprStm( (Expr) yyVals[0+yyTop] ); }
  break;
case 198:
#line 1062 "cs-parser.jay"
  { yyVal = new ExprStm( (Expr) yyVals[0+yyTop] ); }
  break;
case 199:
#line 1063 "cs-parser.jay"
  { yyVal = new ExprStm( (Expr) yyVals[0+yyTop] ); }
  break;
case 200:
#line 1064 "cs-parser.jay"
  { yyVal = new ExprStm( (Expr) yyVals[0+yyTop] ); }
  break;
case 201:
#line 1065 "cs-parser.jay"
  { yyVal = new ExprStm( (Expr) yyVals[0+yyTop] ); }
  break;
case 202:
#line 1066 "cs-parser.jay"
  { yyVal = new ExprStm( (Expr) yyVals[0+yyTop] ); }
  break;
case 204:
#line 1078 "cs-parser.jay"
  { chkb(yyVals[-2+yyTop]); yyVal = new IfStm( (BoolExpr) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], null ); }
  break;
case 205:
#line 1081 "cs-parser.jay"
  { chkb(yyVals[-4+yyTop]); yyVal = new IfStm( (BoolExpr) yyVals[-4+yyTop], (Statement) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop] ); }
  break;
case 209:
#line 1092 "cs-parser.jay"
  { ++loopDepth; }
  break;
case 210:
#line 1094 "cs-parser.jay"
  { chkb(yyVals[-2+yyTop]); yyVal = new WhileStm( (BoolExpr) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop] ); --loopDepth;}
  break;
case 211:
#line 1098 "cs-parser.jay"
  { ++loopDepth; }
  break;
case 212:
#line 1100 "cs-parser.jay"
  { chkb(yyVals[-2+yyTop]); yyVal = new DoStm( (BoolExpr) yyVals[-2+yyTop], (Statement) yyVals[-5+yyTop] ); --loopDepth; }
  break;
case 213:
#line 1106 "cs-parser.jay"
  { currBlock = new Block( currBlock ); ++loopDepth; }
  break;
case 214:
#line 1109 "cs-parser.jay"
  { 
	    currBlock.addStm( new ForStm( (BoolExpr) yyVals[-4+yyTop], (IList) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop] ) );
	    --loopDepth;
	    yyVal = currBlock; 
	    currBlock = currBlock.getParent(); 
	  }
  break;
case 218:
#line 1123 "cs-parser.jay"
  { currBlock.addStm( (IList) yyVals[0+yyTop] ); }
  break;
case 219:
#line 1127 "cs-parser.jay"
  { yyVal = new BoolExpr( true ); }
  break;
case 220:
#line 1128 "cs-parser.jay"
  { chkb( yyVals[0+yyTop] ); yyVal = yyVals[0+yyTop]; }
  break;
case 221:
#line 1132 "cs-parser.jay"
  { yyVal = null; }
  break;
case 223:
#line 1139 "cs-parser.jay"
  { IList l = new ArrayList( 2 ); l.Add( yyVals[0+yyTop] ); yyVal = l; }
  break;
case 224:
#line 1141 "cs-parser.jay"
  { ( (IList) yyVals[-2+yyTop] ).Add( yyVals[0+yyTop] ); yyVal = yyVals[-2+yyTop]; }
  break;
case 230:
#line 1155 "cs-parser.jay"
  { if( loopDepth == 0 ) { errBreakContNoLoop(); yyVal = null; } else yyVal = new BreakStm(); }
  break;
case 231:
#line 1160 "cs-parser.jay"
  { if( loopDepth == 0 ) { errBreakContNoLoop(); yyVal = null; } else yyVal = new ContinueStm(); }
  break;
case 232:
#line 1165 "cs-parser.jay"
  { if( currMethod.returnType() != null ) { errVoidReturn(); yyVal = null; } 
	    else yyVal = new ReturnStm( null ); }
  break;
case 233:
#line 1168 "cs-parser.jay"
  { if( currMethod.returnType() == null ) { errNoVoidReturn(); yyVal = null; } 
	    else if( !( (Expr) yyVals[-1+yyTop] is NullExpr ) && ( (Expr) yyVals[-1+yyTop] ).type() != currMethod.returnType() ) 
		{ errReturnType(); yyVal = null; }
	    else 
		yyVal = new ReturnStm( (Expr) yyVals[-1+yyTop] ); 
	  }
  break;
case 234:
#line 1178 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; }
  break;
case 235:
#line 1182 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 236:
#line 1184 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 237:
#line 1186 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 238:
#line 1191 "cs-parser.jay"
  { yyVal = new ThrowStm ( (Expr) yyVals[-1+yyTop] ); }
  break;
case 239:
#line 1193 "cs-parser.jay"
  { errUnimplOp(); yyVal = null; }
  break;
case 240:
#line 1199 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 241:
#line 1201 "cs-parser.jay"
  { Block b = new Block( currBlock );
	    b.addStm( (Block) yyVals[-3+yyTop] );
	    b.addStm( (Block) yyVals[0+yyTop] );
	    yyVal = b;
	  }
  break;
case 242:
#line 1207 "cs-parser.jay"
  {
		Report.Error (1524, lexer.Location, "Expected catch or finally");
	  }
  break;
case 251:
#line 1237 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; }
  break;
case 252:
#line 1241 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; }
  break;
case 253:
#line 1245 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; }
  break;
case 254:
#line 1262 "cs-parser.jay"
  { 
	    if( ( (int) yyVals[0+yyTop] ) == 0 )
		yyVal = yyVals[-1+yyTop];
	    else {
	      yyVal = currDeclType = arrayType; arrayDeclDim = (int) yyVals[0+yyTop]; arrayElementType = (Type) yyVals[-1+yyTop]; 
	    }
	  }
  break;
case 258:
#line 1284 "cs-parser.jay"
  { if( !( yyVals[0+yyTop] is NullExpr ) && ( (Expr) yyVals[0+yyTop] ).type() != currDeclType )
		errAssignmentTypeMismatch();
	    else {
		if( currBlock != null ) { /* local variable*/
		    if( currDeclType == stubType )
			errStubTypeForbidden();
		    else {
			if( currDeclType == arrayType )
			    currBlock.addVar( new Variable( (string) yyVals[-2+yyTop],
				arrayElementType, stubDeclClass, true, null, false ) ); 
			else
			    currBlock.addVar( new Variable( (string) yyVals[-2+yyTop], currDeclType, 
				stubDeclClass, false, null, false ) ); 
			currBlock.addStm( new ExprStm( makeRefAssign( 
			    new Ref( new QID( (string) yyVals[-2+yyTop] ), false ), (Expr) yyVals[0+yyTop] ) ) );
		    }
		} else { /* class field*/
		    Variable f;
		    if( currDeclType == arrayType )
			f = new Variable( (string) yyVals[-2+yyTop], arrayElementType, stubDeclClass, 
			    true, attributes, modifiers.Contains( Modifiers.PUBLIC ) );
		    else
			f = new Variable( (string) yyVals[-2+yyTop], currDeclType, stubDeclClass, false, 
			    attributes, modifiers.Contains( Modifiers.PUBLIC ) );
		    currClass.addField( f );
		    currClass.addInitStatement( new ExprStm( makeRefAssign( 
			    new Ref( new QID( (string) yyVals[-2+yyTop] ), true ), (Expr) yyVals[0+yyTop] ) ) );
		}
	    }
          }
  break;
case 259:
#line 1316 "cs-parser.jay"
  { if( currBlock != null ) { /* local variable*/
		if( currDeclType == stubType )
		    errStubTypeForbidden();
		else {
		    if( currDeclType == arrayType )
			currBlock.addVar( new Variable( (string) yyVals[0+yyTop],
			    arrayElementType, stubDeclClass, true, null, false ) ); 
		    else
			currBlock.addVar( new Variable( (string) yyVals[0+yyTop], 
			    currDeclType, stubDeclClass, false, null, false ) ); 
		}
	    } else {
		Variable f;
		if( currDeclType == arrayType )
		    f = new Variable( (string) yyVals[0+yyTop], arrayElementType,
			stubDeclClass, true, attributes, modifiers.Contains( Modifiers.PUBLIC ) );
		else
		    f = new Variable( (string) yyVals[0+yyTop], currDeclType, stubDeclClass,
			false, attributes, modifiers.Contains( Modifiers.PUBLIC ) );
		currClass.addField( f );
	    }
	  }
  break;
case 262:
#line 1347 "cs-parser.jay"
  { currMethod.setBody( (Block) yyVals[0+yyTop] ); }
  break;
case 263:
#line 1348 "cs-parser.jay"
  { currClass.delMethod( currMethod ); }
  break;
case 264:
#line 1355 "cs-parser.jay"
  { currMethod = new Method( (string) yyVals[-3+yyTop], (Type) yyVals[-4+yyTop], (IList) yyVals[-1+yyTop], 
		                     (List) yyVals[-6+yyTop], modifiers.Contains( Modifiers.STATIC ) );
	    if( !currClass.addMethod( currMethod ) )
		errDuplicateMethod(); 
	  }
  break;
case 265:
#line 1363 "cs-parser.jay"
  { yyVal = new List(0); }
  break;
case 268:
#line 1375 "cs-parser.jay"
  { IList pars = new List(4); pars.Add (yyVals[0+yyTop]); yyVal = pars; }
  break;
case 269:
#line 1377 "cs-parser.jay"
  { ( (IList) yyVals[-2+yyTop] ).Add(yyVals[0+yyTop]); yyVal = yyVals[-2+yyTop]; }
  break;
case 270:
#line 1382 "cs-parser.jay"
  { 
	    if( currDeclType == arrayType )
		yyVal = new VariableDecl( (string) yyVals[0+yyTop], arrayElementType,
		    stubDeclClass, true, (List) yyVals[-3+yyTop], false ); 
	    else
		yyVal = new VariableDecl( (string) yyVals[0+yyTop], currDeclType, 
		    stubDeclClass, false, (List) yyVals[-3+yyTop], false ); 
	  }
  break;
case 276:
#line 1417 "cs-parser.jay"
  { currClass = new Class( (string) yyVals[-1+yyTop], (List) yyVals[-4+yyTop], (string) yyVals[0+yyTop] ); }
  break;
case 277:
#line 1419 "cs-parser.jay"
  { classes.Add( currClass.getName(), currClass ); }
  break;
case 278:
#line 1424 "cs-parser.jay"
  { errClassBase(); yyVal = ""; }
  break;
case 280:
#line 1441 "cs-parser.jay"
  { QID qid = (QID) yyVals[0+yyTop];
	    if( qid.id == "CCM" )
		qid = qid.next;
	    if( qid.next != null ) {
		errClassBase();
		yyVal = null;
	    } else if( qid.id == "Stub" || qid.id == "StubR" || qid.id == "StubS" )
		yyVal = "Stub";
	    else if( qid.id == "Comp" )
		yyVal = "Comp";
	    else {
		errClassBase();
		yyVal = "";
	    }
	  }
  break;
case 290:
#line 1490 "cs-parser.jay"
  { yyVal = modifiers = new ArrayList(); }
  break;
case 291:
#line 1491 "cs-parser.jay"
  { yyVal = modifiers = (IList) yyVals[0+yyTop]; }
  break;
case 292:
#line 1495 "cs-parser.jay"
  { 
	    IList l = new ArrayList(); 
	    if( yyVals[0+yyTop] != null ) 
		l.Add( yyVals[0+yyTop] ); 
	    yyVal = l; 
	  }
  break;
case 293:
#line 1501 "cs-parser.jay"
  { 
	    if( yyVals[0+yyTop] != null )
		( (IList) yyVals[-1+yyTop] ).Add( yyVals[0+yyTop] ); 
	    yyVal = yyVals[-1+yyTop]; 
	  }
  break;
case 294:
#line 1509 "cs-parser.jay"
  { yyVal = null; }
  break;
case 295:
#line 1510 "cs-parser.jay"
  { yyVal = Modifiers.PUBLIC; }
  break;
case 296:
#line 1511 "cs-parser.jay"
  { yyVal = null; }
  break;
case 297:
#line 1512 "cs-parser.jay"
  { yyVal = null; }
  break;
case 298:
#line 1513 "cs-parser.jay"
  { yyVal = null; }
  break;
case 299:
#line 1514 "cs-parser.jay"
  { yyVal = null; }
  break;
case 300:
#line 1515 "cs-parser.jay"
  { yyVal = null; }
  break;
case 301:
#line 1516 "cs-parser.jay"
  { yyVal = Modifiers.STATIC; }
  break;
case 302:
#line 1517 "cs-parser.jay"
  { yyVal = null; }
  break;
case 303:
#line 1518 "cs-parser.jay"
  { yyVal = null; }
  break;
case 304:
#line 1519 "cs-parser.jay"
  { yyVal = null; }
  break;
case 305:
#line 1520 "cs-parser.jay"
  { yyVal = null; }
  break;
case 306:
#line 1521 "cs-parser.jay"
  { yyVal = null; }
  break;
case 307:
#line 1522 "cs-parser.jay"
  { yyVal = null; }
  break;
case 308:
#line 1526 "cs-parser.jay"
  { currMethod.setBody( (Block) yyVals[0+yyTop] ); }
  break;
case 310:
#line 1533 "cs-parser.jay"
  { if( modifiers.Contains( Modifiers.STATIC ) )
		errStaticCtor();
	    else {
		currMethod = new Method( (string) yyVals[-3+yyTop], null, (IList) yyVals[-1+yyTop], (List) yyVals[-5+yyTop], false );
		currClass.addConstructor( currMethod );
	    }
	  }
  break;
case 311:
#line 1549 "cs-parser.jay"
  { yyVal = attributes = null; }
  break;
case 312:
#line 1550 "cs-parser.jay"
  { yyVal = attributes; }
  break;
case 316:
#line 1564 "cs-parser.jay"
  { attributes = new List(); attributes.Add( currAttr ); }
  break;
case 317:
#line 1565 "cs-parser.jay"
  { attributes.Add( currAttr ); }
  break;
case 318:
#line 1570 "cs-parser.jay"
  { currAttr = new Attribute( yyVals[0+yyTop].ToString(), new List(), new Hashtable() ); }
  break;
case 329:
#line 1597 "cs-parser.jay"
  { 
	    try {
	        Object val = ( (Expr) yyVals[0+yyTop] ).eval();
		currAttr.positional.Add( val );
	    } catch( Exception ) {} /* ignore non-const expr, handled by primary_expression*/
	  }
  break;
case 332:
#line 1611 "cs-parser.jay"
  { 
	    try {
	        Object val = ( (Expr) yyVals[0+yyTop] ).eval();
		currAttr.named.Add( yyVals[-2+yyTop], val );
	    } catch( Exception ) {} /* ignore non-const expr, handled by primary_expression*/
	  }
  break;
#line default
        }
        yyTop -= yyLen[yyN];
        yyState = yyStates[yyTop];
        int yyM = yyLhs[yyN];
        if (yyState == 0 && yyM == 0) {
//t          if (debug != null) debug.shift(0, yyFinal);
          yyState = yyFinal;
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t               debug.lex(yyState, yyToken,yyname(yyToken), yyLex.value());
          }
          if (yyToken == 0) {
//t            if (debug != null) debug.accept(yyVal);
            return yyVal;
          }
          goto yyLoop;
        }
        if (((yyN = yyGindex[yyM]) != 0) && ((yyN += yyState) >= 0)
            && (yyN < yyTable.Length) && (yyCheck[yyN] == yyState))
          yyState = yyTable[yyN];
        else
          yyState = yyDgoto[yyM];
//t        if (debug != null) debug.shift(yyStates[yyTop], yyState);
	 goto yyLoop;
      }
    }
  }

   static  short [] yyLhs  = {              -1,
    0,    0,    2,    2,    1,    1,    3,    3,    6,    6,
    4,    4,    7,    8,    5,    5,   11,   13,   15,   15,
   16,   16,   17,   17,    9,    9,   14,   14,   18,   18,
   19,   19,   20,   20,   20,   21,   21,   21,   21,   21,
   21,   21,   21,   21,   23,   23,   23,   23,   23,   23,
   23,   23,   22,   25,   25,   24,   26,   27,   27,   28,
   28,   29,   29,   29,   29,   29,   29,   29,   29,   29,
   30,   30,   30,   30,   30,   30,   30,   30,   30,   30,
   38,   38,   38,   38,   40,   41,   41,   41,   39,   39,
   39,   39,   39,   39,   39,   44,   44,   44,   44,   45,
   45,   45,   46,   46,   46,   47,   47,   47,   47,   47,
   47,   47,   48,   48,   48,   49,   49,   50,   50,   51,
   51,   52,   52,   53,   53,   54,   54,   55,   55,   55,
   55,   55,   55,   55,   55,   55,   55,   55,   32,   32,
   35,   36,   42,   43,   33,   56,   56,   57,   57,   58,
   58,   58,   37,   37,   61,   59,   64,   60,   60,   60,
   63,   63,   65,   65,   66,   66,   68,   62,   62,   34,
   31,   71,   69,   70,   70,   72,   72,   73,   73,   73,
   74,   75,   75,   75,   75,   75,   75,   75,   75,   75,
   75,   75,   78,   76,   79,   88,   88,   88,   88,   88,
   88,   88,   80,   89,   89,   81,   81,   81,   93,   90,
   94,   91,   96,   92,   95,   95,   99,   99,   97,   97,
   98,   98,  100,  100,   82,   82,   82,   82,   82,  101,
  102,  104,  104,   86,  103,  103,  103,  105,  105,   83,
   83,   83,  107,  107,  106,  106,  108,  109,  109,  110,
   84,   85,   87,  111,   77,  112,  112,  113,  113,   67,
   67,  114,  114,  115,  117,  117,  118,  119,  119,  120,
  121,  121,  122,  122,   10,  126,  123,  124,  124,  127,
  125,  128,  128,  129,  129,  130,  130,  130,  131,  116,
  116,  133,  133,  134,  134,  134,  134,  134,  134,  134,
  134,  134,  134,  134,  134,  134,  134,  132,  132,  135,
   12,   12,  136,  136,  137,  138,  138,  141,  139,  140,
  140,  142,  142,  142,  143,  143,  144,  144,  146,  145,
  145,  147,
  };
   static  short [] yyLen = {           2,
    2,    1,    0,    1,    1,    2,    1,    1,    1,    2,
    1,    1,    5,    3,    1,    1,    5,    4,    0,    1,
    0,    1,    1,    2,    1,    3,    0,    1,    0,    1,
    0,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    2,    0,    1,    1,    3,    0,    1,    1,
    2,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    3,    3,    1,    1,    1,    1,    1,    1,
    1,    2,    2,    1,    4,    1,    2,    2,    1,    2,
    2,    1,    1,    2,    2,    1,    3,    3,    3,    1,
    3,    3,    1,    3,    3,    1,    3,    3,    3,    3,
    3,    3,    1,    3,    3,    1,    3,    1,    3,    1,
    3,    1,    3,    1,    3,    1,    5,    3,    3,    3,
    3,    3,    3,    3,    3,    3,    3,    3,    1,    1,
    2,    2,    2,    2,    4,    0,    1,    1,    3,    1,
    2,    2,    1,    1,    0,    6,    0,    8,    3,    3,
    0,    1,    2,    4,    1,    3,    1,    1,    3,    4,
    1,    0,    4,    0,    1,    1,    2,    1,    1,    1,
    2,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    3,    2,    1,    1,    1,    1,    1,
    1,    1,    1,    5,    7,    1,    1,    1,    0,    6,
    0,    8,    0,   10,    0,    1,    1,    1,    0,    1,
    0,    1,    1,    3,    1,    1,    1,    1,    1,    2,
    2,    2,    3,    5,    3,    4,    3,    3,    2,    3,
    5,    3,    0,    1,    1,    2,    3,    0,    1,    4,
    2,    2,    2,    2,    2,    1,    3,    3,    1,    1,
    1,    2,    2,    7,    0,    1,    1,    1,    3,    4,
    0,    1,    1,    1,    1,    0,    8,    0,    1,    2,
    3,    0,    1,    1,    2,    1,    1,    1,    5,    0,
    1,    1,    2,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    2,    2,    6,
    0,    1,    1,    2,    4,    1,    3,    0,    3,    0,
    3,    1,    3,    1,    0,    1,    1,    3,    1,    1,
    3,    3,
  };
   static  short [] yyDefRed = {            0,
    4,    0,    0,    0,    0,    2,    5,    7,    8,   11,
   12,   15,   16,    0,  275,    0,  313,    0,    0,   25,
    0,    0,  316,    1,    6,  299,  305,  297,    0,  294,
  304,  298,  296,  295,  302,  300,  301,  307,  303,  306,
    0,    0,  292,  314,    0,    0,   14,    0,    0,    0,
    0,    0,  293,    0,   26,    0,  319,  317,  315,    0,
    0,    0,   13,   63,    0,   70,  171,   62,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   64,   65,   66,
   67,   68,   69,    0,    0,   71,    0,    0,  329,   75,
   77,   78,   79,   80,   89,   96,   84,   92,   93,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  139,
  140,  153,  154,    0,  322,    0,    0,  327,  330,    9,
    0,    0,   28,   17,    0,  276,  279,   38,   46,   43,
   39,   41,   40,   49,   51,   36,   45,   47,   37,   50,
   52,   48,   42,    0,    0,   34,   35,   44,   86,    0,
    0,    0,   83,   90,   91,   82,   95,   94,  143,  144,
    0,    0,    0,    0,  141,  142,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  321,    0,
    0,   10,   23,    0,    0,    0,    0,  160,    0,    0,
   53,   56,    0,  159,   74,    0,    0,   88,   87,  332,
    0,    0,  150,    0,    0,  148,  168,    0,  128,  129,
  130,  133,  131,  132,  134,  135,  136,  138,  137,    0,
   97,   99,   98,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  328,    0,  331,   18,   24,    0,    0,  163,
  260,  261,    0,  165,   60,    0,    0,    0,    0,   85,
  152,  151,  145,    0,  170,    0,    0,    0,  286,    0,
    0,    0,  284,  287,  288,    0,  277,    0,    0,   57,
   61,    0,    0,  149,  169,  127,    0,  172,  263,  262,
  281,  285,  309,  308,  166,  164,   55,  157,  156,    0,
    0,    0,    0,    0,    0,    0,  256,    0,    0,    0,
  211,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  209,    0,  193,    0,    0,    0,    0,    0,    0,  201,
  202,  198,    0,  182,    0,    0,  176,  178,  179,  180,
    0,  183,  184,  185,  186,  187,  188,  189,  190,  191,
  192,    0,  203,  206,  207,  208,  225,  226,  227,  228,
  229,    0,  158,  162,    0,    0,  266,    0,  268,    0,
    0,    0,  289,  230,  251,  231,    0,  213,    0,    0,
    0,    0,    0,  232,    0,  239,    0,    0,  252,  253,
    0,    0,  254,  173,  177,  181,  195,    0,    0,  274,
  273,    0,  272,  310,    0,    0,  258,  257,    0,    0,
  167,    0,  237,  235,    0,    0,  233,  238,  242,    0,
    0,    0,  245,    0,  194,    0,  269,  264,    0,  217,
  223,    0,  216,    0,  236,    0,    0,    0,    0,  249,
  246,    0,    0,  270,    0,    0,    0,    0,  234,    0,
  247,  241,    0,    0,  220,    0,  224,    0,   32,    0,
  210,    0,    0,  205,  250,  212,    0,    0,    0,  214,
  };
  protected static  short [] yyDgoto  = {             4,
    5,    6,    7,    8,    9,  121,   10,   11,   85,   12,
   13,   14,   61,  124,  122,  204,  205,   50,  480,  145,
  146,  147,  148,  211,  318,  212,  276,  277,   86,   87,
   88,  223,   90,   91,   92,   93,   94,   95,   96,   97,
  151,   98,   99,  100,  101,  102,  103,  104,  105,  106,
  107,  108,  109,  110,  111,  224,  225,  226,  112,  113,
  213,  228,  383,  323,  272,  273,  274,  432,  354,  355,
  322,  356,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
  375,  376,  411,  397,  452,  430,  476,  487,  453,  454,
  377,  378,  379,  380,  381,  441,  442,  443,  459,  460,
  382,  326,  327,  289,  290,   41,  386,  387,  388,  389,
  422,  423,   15,  126,  269,  207,  127,  291,  292,  293,
  294,  295,   42,   43,  296,   16,   17,   22,   23,   57,
   48,  114,  115,  116,  117,  118,  119,
  };
  protected static  short [] yySindex = {         -253,
    0, -384, -357,    0, -253,    0,    0,    0,    0,    0,
    0,    0,    0, 1903,    0, -269,    0, -246, -304,    0,
 -229, -211,    0,    0,    0,    0,    0,    0, -357,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 -138,   48,    0,    0, -357, -258,    0, -200, -357, -178,
 -271, -225,    0, -244,    0, 2196,    0,    0,    0, -128,
 -140, -133,    0,    0, 1268,    0,    0,    0, 1009, 2217,
 2217, 2217, 2217, 2217, 2217, 2217, 2217,    0,    0,    0,
    0,    0,    0, -124, -144,    0, 3113, -130,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, -162,
 -267, -160, -161, -164, -136, -131, -132, -151, -332,    0,
    0,    0,    0, -113,    0, -111, -110,    0,    0,    0,
 -128, -269,    0,    0, -357,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, -229, -250,    0,    0,    0,    0, -104,
 -299, -312,    0,    0,    0,    0,    0,    0,    0,    0,
 2217, 1896, 2217, 2217,    0,    0, 2217, 2217, 2217, 2217,
 2217, 2217, 2217, 2217, 2217, 2217, -357, 2217, 2217, 2217,
 2217, 2217, 2217, 2217, 1268, 1268, 2217, 2217, 2217, 2217,
 2217, 2217, 2217, 2217, 2217, 2217, 2217, 2217,    0, 2196,
 -184,    0,    0,  -99, -269, -229,  -97,    0, 1923, 1998,
    0,    0, -100,    0,    0, -102, 2217,    0,    0,    0,
 2217, 2217,    0,  -98, -101,    0,    0, -240,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, -229,
    0,    0,    0, -162, -162, -267, -267,  -92,  -92, -160,
 -160, -160, -160, -161, -161, -164, -136, -131, -132,  -96,
 -151, -110,    0, -124,    0,    0,    0, -269, -140,    0,
    0,    0,  -90,    0,    0,  -80,  -81, -205, 1896,    0,
    0,    0,    0, 1896,    0, 2217, 2217,   48,    0, -308,
  -74, -269,    0,    0,    0, -272,    0, 2019,  -71,    0,
    0,  -92,  -73,    0,    0,    0, 1300,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  -70,
 -331,  936,  -63, -269, -280, -125,    0,  -76,  -61,  -65,
    0,  -66, -252,  -56,  -55, 2094, 2121,  -61,  -61,  -61,
    0, 2217,    0,  -58,  -92, 3113,    0,    0,    0,    0,
    0,    0,    0,    0,  -49,  936,    0,    0,    0,    0,
  -50,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  -46,    0,    0,    0,    0,    0,    0,    0,    0,
    0, -114,    0,    0, -219,  -51,    0,  -42,    0, -269,
 2019, -114,    0,    0,    0,    0, 1402,    0, 2217,  -37,
  -35, 2217, 2217,    0,  -34,    0,  -32, -216,    0,    0,
  -31,  936,    0,    0,    0,    0,    0,  -47,  -26,    0,
    0, 1268,    0,    0, -269,  -23,    0,    0,    8, 1110,
    0,  -16,    0,    0,   -9,   -3,    0,    0,    0,    6,
   85,   75,    0, 2217,    0, -319,    0,    0,   13,    0,
    0,    9,    0,   18,    0, 1402, 1402, 1268,  -61,    0,
    0,  -61,   15,    0, 2217, 2217, 2222,   92,    0, -317,
    0,    0, 1402,   22,    0,   19,    0, 1402,    0,   24,
    0,   21, 2222,    0,    0,    0,   26,   18, 1402,    0,
  };
  protected static  short [] yyRindex = {          139,
    0,    0,    0,    0,  139,    0,    0,    0,    0,    0,
    0,    0,    0,  106,    0,  380,    0, -159,    0,    0,
 -172,   31,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1223,    0,    0,    0,    0,    0, -180,   33,    0,
    0,    0,    0,    0,    0,   32,    0,    0,    0,  673,
    5,   39,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1788, 1435,    0,  798, 1508,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2301,
 2430, 2560,  649, 2039, 3192,  -18,  -87,  307, -208,    0,
    0,    0,    0,    0,    0,   37,   38,    0,    0,    0,
 2905, 2948,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  137,   45,    0,    0,    0,    0,    0,
    0,  798,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   46,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2991,   50,    0,    0,    0,   51,
    0,    0,    0,    0,    0,   51,    0,    0,    0,    0,
    0,    0,    0,    0,   54,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1581,
    0,    0,    0, 2345, 2388, 2473, 2517, 2602, 2645, 2689,
 2732, 2774, 2817,  969, 1391, 3169, 3215,  180,  382,    0,
  316,   55,    0,    0,    0,    0,    0,  455,    5,    0,
    0,    0,   60,    0,    0,    0, -178,    0,   46,    0,
    0,    0,    0,    0,    0,    0,    0, 1344,    0,    0,
    0,  530,    0,    0,    0,    0,    0,   61,    0,    0,
    0, 1654,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, -320,
    0,   62, 1727, -245, -122,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 3080,  -22,    0, 2956, 2989, 3018,    0,
    0,    0, 3051,    0,    0,   65,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1384,    0,    0,   63,    0, -245,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  115,    0,    0,
    0,    0,    0,    0,    0,    0,    0, -122,   58,    0,
    0,    0,    0,    0, 1183,    0,    0,    0,    0,   66,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   70,
  616,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   67,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   69,    0,  776,    0,   74,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   76,    0,    0,    0,    0,   78,    0,    0,
  };
  protected static  short [] yyGindex = {            0,
    0,  413,  414,  -36,  -94,    0,    0,    0,   -2,    0,
    0, -222,    0,  162,    0,    0,    0,  159,    0, -166,
  -67,    0,    0, -268,   89,  284,    0,    0,    0,  -62,
    0,  -11, -243,    0, -185,   82,    0,    0,  800,    0,
    0,  107,  147,  -78,   -7,    3,    4,  246,  249,  245,
  248,  247,    0,    0,  149,  167,    0,  164,  161,    0,
    0,  243,    0,    0, -142,    0, -291,    0, -274,    0,
    0,    0, -314,    0, -231,    0,   28,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, -394,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  -24,
    0,    0,    0,    0,    0,    0,    0,   23,    0,    0,
    0,   80,   68,    0,    0,  177,   77,    0,    0,   41,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  176,
    0,    0,    0,  428,    0,    0,  456,    0,  425,    0,
    0,    0,    0,    0,  275,  277,  279,
  };
  protected static  short [] yyTable = {            19,
   21,  149,  214,    1,   27,  208,  315,  152,  152,  152,
  152,  152,  152,  152,  152,  310,  399,  216,  248,  249,
  311,  314,  311,  120,  400,  311,   51,  203,   25,  216,
  311,  216,   25,  317,  311,  451,  163,  197,  308,  439,
   18,  415,   54,  311,   89,  288,   21,  309,   46,  216,
  311,   47,  217,  440,  395,  311,  198,  150,  165,  311,
  166,  311,  144,  408,  409,  410,  218,   20,  311,  288,
  390,  311,  477,  311,  308,   60,  317,  311,  347,    3,
  391,   46,    2,  313,  202,  311,  311,  420,  451,  311,
  181,  182,  311,  325,  421,    3,  209,  445,  210,  427,
  185,  385,  244,  245,   25,  464,  265,  479,   46,  285,
  267,   63,  347,  286,   45,  152,  152,  152,  152,  152,
  152,  152,  206,   46,  152,  152,  152,  152,  152,  152,
  152,  152,  152,  152,   52,  152,  348,  186,    3,  126,
  321,  126,   49,  126,  302,  126,  126,  126,  286,  220,
   56,  227,  229,  347,  152,  230,  231,  232,  233,  234,
  235,  236,  237,  238,  239,  429,   55,  385,  347,  320,
  348,   59,  401,  320,  240,  246,  247,  318,  318,  311,
  384,  318,  144,  144,  471,  260,  347,  472,   89,  250,
  251,  252,  253,   25,  254,  255,   25,  271,  227,   62,
  187,  188,  385,  178,  179,  180,  162,    2,   46,  281,
  282,  348,  347,  347,  183,  123,  184,  189,  191,  190,
  192,  125,  177,  347,  468,  469,  348,  193,  392,  347,
  393,  259,  195,  259,  347,  196,  161,  194,  199,  347,
  264,  481,  200,  201,  348,  347,  484,  215,  266,  268,
  279,  275,  284,  283,  345,  446,  216,  490,  287,  346,
  122,   27,  122,  298,  122,   27,  122,  122,  122,  300,
  348,  348,  301,  311,  305,  306,  316,   27,  319,  394,
  324,  348,  122,  209,  398,  308,  271,  348,  345,   27,
  396,  470,  348,  346,  402,  403,  412,  348,  414,  122,
  424,  122,   27,  348,  144,  416,   27,   27,   26,  417,
  418,  425,   27,  391,   27,   27,   27,   27,  433,  444,
  434,  437,   27,  438,  405,  407,   27,  392,  448,  120,
  150,  120,   27,  120,  346,  120,  120,  120,   27,  455,
   27,   27,  456,   27,  345,   28,  120,  449,  457,  346,
   30,  120,   27,   27,  440,   31,  458,   32,   33,   34,
   35,  462,  345,  465,  466,   36,  473,  346,  120,   37,
  120,  467,  478,  482,  483,  485,  486,  489,  290,  271,
   29,   38,   30,  325,   39,  278,   40,  431,  326,  324,
  435,  436,   33,  346,  346,  155,  280,  146,   33,  311,
   58,  243,   54,  349,  346,  147,  323,   29,   30,  174,
  346,  311,  175,  255,  267,  346,  248,   24,   25,  144,
  346,  215,  218,  311,  219,   31,  346,  221,  350,  222,
  297,  299,  463,  413,  219,   33,  311,  349,  256,  258,
  311,  311,  257,  259,  261,  303,  311,  304,  311,  311,
  311,  311,  278,  474,  475,  144,  311,  450,  488,  428,
  311,  419,  350,  461,  307,  447,  426,  312,  351,   53,
  352,   44,  311,   58,  262,  311,  263,  311,  349,  265,
    0,    0,  353,   33,   33,   33,   33,   33,   33,    0,
   33,   33,   33,  349,    0,    0,    0,    0,   33,   33,
   33,   33,  351,  350,  352,   33,   33,    0,    0,    0,
    0,  349,    0,    0,    0,   33,  353,   33,  350,   33,
    0,   33,    0,   33,    0,   33,    0,  121,    0,  121,
    0,  121,    0,  121,  121,  121,  350,  349,  349,    0,
    0,    0,    0,  351,  121,  352,    0,    0,  349,  121,
    0,    0,    0,    0,  349,    0,    0,  353,  351,  349,
  352,   33,  350,  350,  349,    0,  121,    0,  121,    0,
  349,    0,  353,  350,    0,    0,  351,    0,  352,  350,
    0,    0,    0,    0,  350,    0,    0,    0,    0,  350,
  353,    0,    0,    0,    0,  350,    0,    0,    0,    0,
    0,    0,  351,  351,  352,  352,    0,    0,    0,    0,
    0,    0,    0,  351,    0,  352,  353,  353,    0,  351,
    0,  352,    0,    0,  351,    0,  352,  353,    0,  351,
    0,  352,    0,  353,    0,  351,    0,  352,  353,    0,
  312,    0,    0,  353,    0,  312,    0,  312,    0,  353,
  312,    0,  312,    0,  124,  312,  124,    0,  124,  312,
  124,  124,  124,  125,  312,  125,    0,  125,  312,  125,
  125,  125,    0,    0,    0,  312,  124,  312,    0,    0,
  312,  312,  312,    0,  312,  125,  312,  312,    0,  312,
  312,  312,  312,  312,    0,  124,  312,  312,  312,    0,
    0,  312,  312,    0,  125,    0,    0,    0,    0,    0,
  312,  312,    0,  312,  312,  311,  312,  312,  312,    0,
  311,    0,  311,    0,    0,  311,    0,    0,    0,  123,
  311,  123,    0,  123,  311,  123,  123,  123,    0,  311,
    0,    0,    0,  311,    0,    0,    0,    0,    0,    0,
  311,  123,  311,    0,    0,  311,    0,  311,    0,  311,
    0,    0,  311,    0,  311,  311,  311,  311,  123,    0,
  123,  311,  311,  311,    0,    0,  311,  311,    0,    0,
    0,    0,    0,    0,    0,  311,  311,    0,  311,  311,
  311,  311,  311,  311,    0,  311,    0,  311,    0,    0,
  311,    0,  282,    0,  312,  311,    0,    0,    0,  311,
    0,    0,    0,    0,  311,    0,    0,    0,  311,    0,
    0,    0,    0,    0,    0,  311,    0,  311,    0,    0,
  311,    0,  311,    0,  311,    0,    0,  311,    0,  311,
  311,  311,  311,    0,    0,    0,  311,  311,  311,    0,
    0,  311,  311,    0,    0,    0,    0,    0,    0,    0,
  311,  311,    0,  311,  311,    0,  311,  311,  311,  153,
  154,  155,  156,  157,  158,  159,  160,  283,    0,  311,
    0,  240,  240,  240,    0,    0,  240,  240,    0,    0,
  240,  240,    0,    0,  240,  240,  240,    0,    0,    0,
    0,  240,  244,    0,  240,  240,    0,  240,  240,    0,
    0,  240,    0,    0,    0,  240,  240,    0,  240,  240,
  240,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  240,    0,  240,   19,  240,    0,    0,    0,  240,    0,
    0,  240,  240,  240,  240,   19,  240,  240,  240,  240,
  240,    0,    0,  240,  311,  240,    0,   19,    0,    0,
    0,    0,  240,  240,    0,    0,  240,    0,    0,    0,
   19,  240,    0,    0,   19,   19,    0,  241,  242,  243,
   19,    0,   19,   19,   19,   19,  240,    0,  240,    0,
   19,    0,    0,    0,   19,    0,  113,    0,  113,    0,
  113,    0,  113,  113,  113,    0,   19,    0,    0,   19,
    0,   19,  113,  113,    0,    0,  280,  113,  113,    0,
   19,   19,    0,    0,    0,    0,    0,    0,  240,    0,
  240,  113,  240,  113,  240,  113,  240,  113,  240,    0,
  240,  204,  204,  204,    0,    0,  204,  204,    0,    0,
  204,  204,    0,    0,  204,  204,    0,    0,    0,   81,
    0,  204,    0,    0,  204,  204,    0,  204,  204,    0,
    0,  204,    0,    0,    0,  204,  204,    0,  204,  204,
  204,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  204,    0,  204,    0,  204,    0,   81,    0,  204,    0,
    0,  204,  204,  204,  204,    0,  204,  204,  204,  204,
  204,    0,    0,  204,    0,  204,    0,    0,    0,    0,
    0,    0,  204,  204,    0,    0,  204,    0,    0,    0,
    0,  204,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   81,  204,   81,  204,   81,
    0,   81,   81,   81,    0,   81,   81,    0,    0,   81,
   81,   81,   81,   81,   81,   81,   81,   81,    0,    0,
    0,    0,   81,    0,   81,    0,   81,    0,   81,    0,
   81,    0,   81,    0,   81,    0,   81,    0,  204,    0,
  204,    0,  204,    0,  204,    0,  204,    0,  204,    0,
  204,  128,  328,  129,    0,    0,  130,  329,    0,    0,
  330,  131,    0,    0,  331,  132,    0,    0,    0,    0,
    0,   64,    0,    0,  133,  332,    0,  333,  334,    0,
    0,  134,    0,    0,    0,  335,  135,    0,   65,   66,
  136,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  336,    0,  137,    0,  138,    0,    0,    0,  139,    0,
    0,   67,  337,   68,  338,    0,  140,  141,  339,  340,
  142,    0,    0,  143,  128,  341,  129,    0,    0,  130,
    0,    0,  308,    0,  131,    0,  342,    0,  132,    0,
    0,  343,    0,    0,   64,    0,    0,  133,    0,    0,
    0,    0,    0,    0,  134,    0,   76,    0,   77,  135,
    0,   65,   66,  136,    0,    0,  114,    0,  114,    0,
  114,    0,  114,  114,  114,  137,    0,  138,    0,    0,
    0,  139,  114,  114,   67,    0,   68,  114,  114,  140,
  141,    0,    0,  142,    0,    0,  143,    0,   78,    0,
   79,  114,   80,  114,   81,  114,   82,  114,   83,   69,
  344,    0,    0,    0,    0,   70,   71,   72,   73,    0,
    0,    0,   74,    0,   75,  128,    0,  129,    0,   76,
  130,   77,    0,    0,    0,  131,    0,    0,    0,  132,
    0,    0,    0,    0,    0,   64,    0,    0,  133,    0,
    0,    0,    0,    0,    0,  134,    0,    0,    0,    0,
  135,    0,   65,   66,  136,    0,    0,    0,    0,    0,
    0,   78,    0,   79,    0,   80,  137,   81,  138,   82,
    0,   83,  139,   20,    0,   67,    0,   68,    0,    0,
  140,  141,    0,    0,  142,    0,    0,  143,  311,    0,
  311,    0,    0,  311,    0,    0,    0,    0,  311,    0,
  342,    0,  311,    0,    0,    0,    0,    0,    0,    0,
    0,  311,    0,    0,    0,    0,    0,    0,  311,    0,
   76,    0,   77,  311,    0,    0,    0,  311,  291,  311,
  291,    0,    0,  291,    0,  291,  311,    0,  291,  311,
    0,  311,  291,    0,    0,  311,    0,    0,    0,    0,
    0,  291,    0,  311,  311,    0,    0,  311,  291,    0,
  311,    0,   78,  291,   79,    0,   80,  291,   81,    0,
   82,    0,   83,  128,   20,  129,    0,    0,  130,  291,
    0,  291,    0,  131,    0,  291,    0,  132,    0,    0,
    0,    0,    0,  291,  291,    0,  133,  291,    0,    0,
  291,    0,    0,  134,    0,  128,    0,  129,  135,    0,
  130,    0,  136,    0,    0,  131,    0,    0,    0,  132,
    0,    0,    0,    0,  137,    0,  138,    0,  133,    0,
  139,    0,    0,    0,    0,  134,    0,    0,  140,  141,
  135,    0,  142,    0,  136,  143,    0,  311,    0,  290,
    0,  290,    0,    0,  290,    0,  137,    0,  138,  290,
    0,    0,  139,  290,    0,    0,    0,    0,    0,    0,
  140,  141,  290,    0,  142,    0,    0,  143,    0,  290,
    0,    0,    0,    0,  290,    0,    0,  291,  290,  271,
    0,  271,    0,    0,  271,    0,    0,    0,    0,  271,
  290,    0,  290,  271,    0,    0,  290,    0,  328,    0,
    0,    0,  271,  329,  290,  290,  330,    0,  290,  271,
  331,  290,    0,    0,  271,    0,    0,   64,  271,    0,
    0,  332,   20,  333,  334,    0,   72,    0,    0,    0,
  271,  335,  271,    0,   65,   66,  271,    0,    0,    0,
    0,    0,    0,    0,  271,  271,  336,    0,  271,    0,
    0,  271,    0,    0,  320,    0,    0,   67,  337,   68,
  338,    0,    0,   72,  339,  340,    0,    0,  115,    0,
  115,  341,  115,    0,  115,  115,  115,    0,  308,    0,
    0,    0,  342,    0,  115,  115,    0,  343,    0,  115,
  115,    0,    0,    0,    0,    0,    0,    0,  290,   76,
    0,    0,   76,  115,   77,  115,    0,  115,    0,  115,
    0,    0,   72,   72,   72,    0,   72,    0,   72,   72,
   72,    0,   72,   72,    0,   72,   72,   72,   72,   72,
   72,   72,   72,   72,   72,   72,   76,   72,  271,   72,
    0,   72,    0,   72,   78,   72,   79,   72,   80,   72,
   81,   72,   82,   72,   83,   72,   20,   72,    0,   72,
    0,   72,    0,   72,    0,   72,    0,   72,    0,   72,
    0,   72,   73,   72,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   76,   76,   76,    0,   76,
    0,   76,   76,   76,    0,   76,   76,    0,   76,   76,
   76,   76,   76,   76,   76,   76,   76,   76,   76,   73,
   76,    0,   76,    0,   76,    0,   76,    0,   76,    0,
   76,    0,   76,    0,   76,    0,   76,    0,   76,    0,
   76,    0,   76,    0,   76,    0,   76,    0,   76,    0,
   76,    0,   76,    0,   76,   54,   76,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   73,   73,
   73,    0,   73,    0,   73,   73,   73,    0,   73,   73,
    0,   73,   73,   73,   73,   73,   73,   73,   73,   73,
   73,   73,   54,   73,    0,   73,    0,   73,    0,   73,
    0,   73,    0,   73,    0,   73,    0,   73,    0,   73,
    0,   73,    0,   73,    0,   73,    0,   73,    0,   73,
    0,   73,    0,   73,    0,   73,    0,   73,  161,   73,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   54,   54,    0,   54,    0,   54,    0,   54,   54,   54,
    0,   54,   54,    0,   54,   54,   54,   54,   54,   54,
   54,   54,   54,   54,   54,  161,   54,    0,   54,    0,
   54,    0,   54,    0,   54,    0,   54,    0,   54,    0,
   54,    0,   54,    0,   54,    0,   54,    0,   54,   25,
   54,    0,   54,    0,   54,    0,   54,    0,   54,    0,
   54,    0,   54,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  161,  161,  161,    0,  161,    0,
  161,  161,  161,    0,  161,  161,   25,  161,  161,  161,
  161,  161,  161,  161,  161,  161,  161,  161,    0,  161,
    0,  161,    0,  161,    0,  161,    0,  161,    0,  161,
    0,  161,    0,  161,    0,  161,    0,  161,    0,  161,
    0,  161,    0,  161,    0,  161,    0,  161,    0,  161,
    0,  161,    0,  161,    0,  161,   25,    0,   25,   25,
   25,   25,    0,    0,    0,   25,   25,    0,    0,   25,
   25,   25,   25,   25,   25,   25,   25,   25,   25,    0,
   25,    0,   25,   26,   25,    0,   25,    0,   25,    0,
   25,    0,   25,    0,   25,    0,   25,    0,   25,    0,
   25,   64,   25,    0,   25,    0,   25,   27,   25,    0,
   25,    0,   25,    0,   25,    0,   25,    0,   65,   66,
   28,    0,  221,    0,   29,   30,    0,    0,   64,  222,
   31,    0,   32,   33,   34,   35,    0,    0,    0,    0,
   36,   67,    0,   68,   37,   65,   66,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   38,    0,    0,   39,
    0,   40,    0,    0,    0,    0,   69,    0,   67,    0,
   68,    0,   70,   71,   72,   73,    0,    0,    0,   74,
    0,   75,    0,    0,    0,    0,   76,    0,   77,  209,
  270,    0,    0,   69,    0,    0,    0,    0,    0,   70,
   71,   72,   73,   64,    0,    0,   74,    0,   75,    0,
    0,    0,    0,   76,    0,   77,    0,    0,    0,    0,
   65,   66,    0,    0,   64,    0,    0,    0,   78,    0,
   79,    0,   80,    0,   81,    0,   82,    0,   83,    0,
   20,   65,   66,   67,    0,   68,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   78,    0,   79,    0,   80,
    0,   81,    0,   82,   67,   83,   68,   20,   69,    0,
    0,  275,    0,    0,   70,   71,   72,   73,    0,    0,
    0,   74,    0,   75,    0,  209,    0,    0,   76,   69,
   77,    0,    0,    0,    0,   70,   71,   72,   73,   64,
    0,    0,   74,    0,   75,    0,  116,    0,  116,   76,
  116,   77,  116,  116,  116,    0,   65,   66,    0,    0,
    0,    0,  116,  116,    0,    0,   64,  116,  116,    0,
   78,    0,   79,    0,   80,    0,   81,    0,   82,   67,
   83,   68,   20,   65,   66,  116,    0,  116,    0,    0,
    0,   78,    0,   79,    0,   80,    0,   81,    0,   82,
    0,   83,    0,   20,   69,    0,   67,    0,   68,  404,
   70,   71,   72,   73,    0,    0,    0,   74,    0,   75,
    0,    0,    0,    0,   76,    0,   77,    0,    0,    0,
    0,   69,    0,    0,    0,    0,  406,   70,   71,   72,
   73,   64,    0,    0,   74,    0,   75,    0,    0,    0,
    0,   76,    0,   77,    0,    0,    0,    0,   65,   66,
    0,    0,   64,    0,    0,    0,   78,   64,   79,    0,
   80,    0,   81,    0,   82,    0,   83,    0,   20,   65,
   66,   67,    0,   68,   65,   66,    0,    0,    0,    0,
    0,    0,    0,   78,    0,   79,    0,   80,    0,   81,
    0,   82,   67,   83,   68,   20,   69,   67,    0,   68,
    0,    0,   70,   71,   72,   73,    0,    0,    0,   74,
    0,   75,  100,    0,    0,    0,   76,   69,   77,    0,
    0,    0,  342,   70,   71,   72,   73,    0,    0,    0,
   74,    0,   75,    0,    0,    0,    0,   76,    0,   77,
    0,    0,   76,    0,   77,    0,    0,    0,    0,  100,
    0,    0,    0,    0,    0,    0,  101,    0,   78,    0,
   79,    0,   80,    0,   81,    0,   82,    0,   83,    0,
   84,    0,    0,    0,    0,    0,    0,    0,    0,   78,
    0,   79,    0,   80,   78,   81,   79,   82,   80,   83,
   81,   20,   82,  101,   83,    0,   20,    0,  100,  102,
  100,    0,  100,    0,  100,  100,  100,    0,  100,  100,
    0,    0,  100,  100,  100,  100,    0,    0,    0,  100,
  100,    0,    0,    0,    0,  100,    0,  100,    0,  100,
    0,  100,    0,  100,    0,  100,  102,  100,    0,  100,
    0,  103,  101,    0,  101,    0,  101,    0,  101,  101,
  101,    0,  101,  101,    0,    0,  101,  101,  101,  101,
    0,    0,    0,  101,  101,    0,    0,    0,    0,  101,
    0,  101,    0,  101,    0,  101,    0,  101,  103,  101,
    0,  101,    0,  101,  104,  102,    0,  102,    0,  102,
    0,  102,  102,  102,    0,  102,  102,    0,    0,  102,
  102,  102,  102,    0,    0,    0,  102,  102,    0,    0,
    0,    0,  102,    0,  102,    0,  102,    0,  102,    0,
  102,  104,  102,    0,  102,    0,  102,  103,  105,  103,
    0,  103,    0,  103,  103,  103,    0,    0,    0,    0,
    0,  103,  103,  103,  103,    0,    0,    0,  103,  103,
    0,    0,    0,    0,  103,    0,  103,    0,  103,    0,
  103,    0,  103,    0,  103,  105,  103,    0,  103,    0,
  104,  106,  104,    0,  104,    0,  104,  104,  104,    0,
    0,    0,    0,    0,  104,  104,  104,  104,    0,    0,
    0,  104,  104,    0,    0,    0,    0,  104,    0,  104,
    0,  104,    0,  104,    0,  104,    0,  104,  106,  104,
    0,  104,    0,  112,  105,    0,  105,    0,  105,    0,
  105,  105,  105,    0,    0,    0,    0,    0,  105,  105,
  105,  105,    0,    0,    0,  105,  105,    0,    0,    0,
    0,  105,    0,  105,    0,  105,    0,  105,    0,  105,
  112,  105,    0,  105,    0,  105,  111,  106,    0,  106,
    0,  106,    0,  106,  106,  106,    0,    0,    0,    0,
    0,  106,  106,  106,  106,    0,    0,    0,  106,  106,
    0,    0,    0,    0,    0,    0,    0,    0,  106,    0,
  106,    0,  106,  111,  106,    0,  106,    0,  106,  112,
  107,  112,    0,  112,    0,  112,  112,  112,    0,    0,
    0,    0,    0,  112,  112,  112,  112,    0,    0,    0,
  112,  112,    0,    0,    0,    0,    0,    0,    0,    0,
  112,    0,  112,    0,  112,    0,  112,  107,  112,    0,
  112,    0,  111,  108,  111,    0,  111,    0,  111,  111,
  111,    0,    0,    0,    0,    0,  111,  111,  111,  111,
    0,    0,    0,  111,  111,    0,    0,    0,    0,    0,
    0,    0,    0,  111,    0,  111,    0,  111,    0,  111,
  108,  111,    0,  111,    0,  109,  107,    0,  107,    0,
  107,    0,  107,  107,  107,    0,    0,    0,    0,    0,
  107,  107,  107,  107,    0,    0,    0,  107,  107,    0,
    0,    0,    0,    0,    0,    0,    0,  107,    0,  107,
    0,  107,  109,  107,    0,  107,    0,  107,  110,  108,
    0,  108,    0,  108,    0,  108,  108,  108,    0,    0,
    0,    0,    0,  108,  108,  108,  108,    0,    0,    0,
  108,  108,    0,    0,    0,    0,    0,    0,    0,    0,
  108,    0,  108,    0,  108,  110,  108,    0,  108,    0,
  108,  109,    0,  109,    0,  109,    0,  109,  109,  109,
    0,    0,    0,    0,    0,  109,  109,  109,  109,    0,
    0,    0,  109,  109,    0,    0,    0,    0,    0,    0,
    0,    0,  109,    0,  109,    0,  109,    0,  109,    0,
  109,    0,  109,    0,  110,   20,  110,    0,  110,    0,
  110,  110,  110,    0,    0,    0,    0,   20,  110,  110,
  110,  110,    0,    0,    0,  110,  110,    0,    0,   20,
    0,    0,    0,    0,    0,  110,    0,  110,    0,  110,
    0,  110,   20,  110,    0,  110,   20,   20,  311,    0,
    0,    0,   20,    0,   20,   20,   20,   20,    0,    0,
  311,    0,   20,    0,    0,    0,   20,    0,    0,    0,
    0,    0,  311,    0,    0,    0,    0,    0,   20,    0,
    0,   20,    0,   20,    0,  311,    0,    0,    0,  311,
  311,  311,   20,   20,    0,  311,    0,  311,  311,  311,
  311,    0,    0,  311,    0,  311,    0,    0,    0,  311,
    0,    0,    0,    0,    0,  311,    0,    0,    0,    0,
    0,  311,    0,    0,  311,    0,  311,    0,  311,    0,
    0,    0,  311,  311,    0,   21,    0,    0,  311,    0,
  311,  311,  311,  311,   75,    0,    0,  196,  311,  196,
    0,  196,  311,    0,    0,    0,   75,    0,    0,    0,
    0,    0,    0,    0,  311,    0,   75,  311,   75,  311,
    0,    0,    0,    0,    0,    0,    0,   78,   22,    0,
  199,    0,  199,    0,  199,    0,   75,    0,   75,   78,
   75,    0,   75,    0,   75,    0,   75,    0,   75,   78,
   75,   78,   75,    0,   75,    0,   79,    0,    0,  200,
    0,  200,    0,  200,    0,    0,    0,    0,   79,   78,
    0,   78,    0,   78,    0,   78,    0,   78,   79,   78,
   79,   78,    0,   78,    0,   78,    0,   78,    0,  153,
    0,    0,  197,    0,  197,    0,  197,    0,   79,    0,
   79,  153,   79,    0,   79,    0,   79,    0,   79,    0,
   79,  153,   79,  153,   79,    0,   79,    0,   25,    0,
   25,    0,   25,    0,    0,    0,    0,    0,    0,    0,
   25,  153,    0,  153,    0,  153,    0,  153,    0,  153,
   25,  153,   25,  153,    0,  153,    0,  153,    0,  153,
    0,  163,    0,    0,    0,    0,    0,    0,    0,    0,
   25,    0,   25,  164,   25,    0,   25,    0,   25,    0,
   25,    0,   25,  165,   25,  166,   25,    0,   25,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  167,    0,  168,    0,  169,    0,  170,
    0,  171,    0,  172,    0,  173,  117,  174,  117,  175,
  117,  176,  117,  117,  117,    0,    0,    0,    0,    0,
    0,    0,  117,  117,    0,    0,    0,  117,  117,  118,
    0,  118,    0,  118,    0,  118,  118,  118,    0,    0,
    0,    0,    0,    0,    0,  117,  118,  117,    0,    0,
  118,  118,  119,    0,  119,    0,  119,    0,  119,  119,
  119,    0,    0,    0,    0,    0,    0,    0,  118,  119,
  118,    0,    0,  119,  119,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  119,    0,  119,
  };
  protected static  short [] yyCheck = {             2,
    3,   69,  145,  257,    0,  256,  298,   70,   71,   72,
   73,   74,   75,   76,   77,  290,  269,  349,  185,  186,
  266,  296,  268,   60,  277,  271,   29,  122,  349,  349,
  276,  349,  353,  302,  280,  430,  349,  370,  347,  256,
  425,  356,   45,  289,   56,  268,   49,  356,  353,  349,
  296,  356,  352,  270,  329,  301,  389,   69,  371,  305,
  373,  307,   65,  338,  339,  340,  366,  425,  314,  292,
  351,  317,  467,  319,  347,  347,  345,  323,  322,  349,
  361,  353,  336,  356,  121,  331,  332,  307,  483,  335,
  358,  359,  338,  425,  314,  349,  347,  412,  349,  391,
  262,  324,  181,  182,  425,  425,  352,  425,  353,  350,
  205,  356,  356,  354,  361,  178,  179,  180,  181,  182,
  183,  184,  125,  353,  187,  188,  189,  190,  191,  192,
  193,  194,  195,  196,  273,  198,  322,  299,    0,  348,
  307,  350,  354,  352,  350,  354,  355,  356,  354,  161,
  351,  163,  164,  397,  217,  167,  168,  169,  170,  171,
  172,  173,  174,  175,  176,  397,  425,  390,  412,  350,
  356,  350,  425,  354,  177,  183,  184,  350,  351,  425,
  323,  354,  185,  186,  459,  197,  430,  462,  200,  187,
  188,  189,  190,  353,  191,  192,  356,  209,  210,  425,
  362,  363,  425,  366,  367,  368,  351,  336,  353,  221,
  222,  397,  456,  457,  375,  356,  377,  379,  383,  381,
  385,  355,  353,  467,  456,  457,  412,  364,  354,  473,
  356,  354,  365,  356,  478,  387,  361,  369,  352,  483,
  425,  473,  354,  354,  430,  489,  478,  352,  348,  347,
  351,  354,  354,  352,  322,  422,  349,  489,  355,  322,
  348,  257,  350,  354,  352,  261,  354,  355,  356,  350,
  456,  457,  354,  348,  286,  287,  348,  273,  352,  356,
  351,  467,  370,  347,  351,  347,  298,  473,  356,  285,
  356,  458,  478,  356,  351,  351,  355,  483,  348,  387,
  352,  389,  298,  489,  307,  356,  302,  303,  261,  356,
  425,  354,  308,  361,  310,  311,  312,  313,  356,  351,
  356,  356,  318,  356,  336,  337,  322,  354,  352,  348,
  342,  350,  285,  352,  397,  354,  355,  356,  334,  356,
  336,  337,  352,  339,  412,  298,  365,  340,  352,  412,
  303,  370,  348,  349,  270,  308,  351,  310,  311,  312,
  313,  287,  430,  351,  356,  318,  352,  430,  387,  322,
  389,  354,  281,  352,  356,  352,  356,  352,  273,  391,
  350,  334,  350,  352,  337,  347,  339,  399,  352,  352,
  402,  403,  256,  456,  457,  351,  347,  352,  262,  261,
  350,  287,  425,  322,  467,  352,  352,  348,  348,  348,
  473,  273,  348,  356,  352,  478,  347,    5,    5,  422,
  483,  356,  356,  285,  356,  352,  489,  352,  322,  352,
  269,  273,  444,  345,  151,  299,  298,  356,  193,  195,
  302,  303,  194,  196,  198,  279,  308,  284,  310,  311,
  312,  313,  210,  465,  466,  458,  318,  430,  483,  392,
  322,  382,  356,  441,  288,  425,  390,  292,  322,   42,
  322,   16,  334,   49,  200,  337,  200,  339,  397,  201,
   -1,   -1,  322,  347,  348,  349,  350,  351,  352,   -1,
  354,  355,  356,  412,   -1,   -1,   -1,   -1,  362,  363,
  364,  365,  356,  397,  356,  369,  370,   -1,   -1,   -1,
   -1,  430,   -1,   -1,   -1,  379,  356,  381,  412,  383,
   -1,  385,   -1,  387,   -1,  389,   -1,  348,   -1,  350,
   -1,  352,   -1,  354,  355,  356,  430,  456,  457,   -1,
   -1,   -1,   -1,  397,  365,  397,   -1,   -1,  467,  370,
   -1,   -1,   -1,   -1,  473,   -1,   -1,  397,  412,  478,
  412,  425,  456,  457,  483,   -1,  387,   -1,  389,   -1,
  489,   -1,  412,  467,   -1,   -1,  430,   -1,  430,  473,
   -1,   -1,   -1,   -1,  478,   -1,   -1,   -1,   -1,  483,
  430,   -1,   -1,   -1,   -1,  489,   -1,   -1,   -1,   -1,
   -1,   -1,  456,  457,  456,  457,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  467,   -1,  467,  456,  457,   -1,  473,
   -1,  473,   -1,   -1,  478,   -1,  478,  467,   -1,  483,
   -1,  483,   -1,  473,   -1,  489,   -1,  489,  478,   -1,
  261,   -1,   -1,  483,   -1,  266,   -1,  268,   -1,  489,
  271,   -1,  273,   -1,  348,  276,  350,   -1,  352,  280,
  354,  355,  356,  348,  285,  350,   -1,  352,  289,  354,
  355,  356,   -1,   -1,   -1,  296,  370,  298,   -1,   -1,
  301,  302,  303,   -1,  305,  370,  307,  308,   -1,  310,
  311,  312,  313,  314,   -1,  389,  317,  318,  319,   -1,
   -1,  322,  323,   -1,  389,   -1,   -1,   -1,   -1,   -1,
  331,  332,   -1,  334,  335,  261,  337,  338,  339,   -1,
  266,   -1,  268,   -1,   -1,  271,   -1,   -1,   -1,  348,
  276,  350,   -1,  352,  280,  354,  355,  356,   -1,  285,
   -1,   -1,   -1,  289,   -1,   -1,   -1,   -1,   -1,   -1,
  296,  370,  298,   -1,   -1,  301,   -1,  303,   -1,  305,
   -1,   -1,  308,   -1,  310,  311,  312,  313,  387,   -1,
  389,  317,  318,  319,   -1,   -1,  322,  323,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  331,  332,   -1,  334,  335,
  261,  337,  338,  339,   -1,  266,   -1,  268,   -1,   -1,
  271,   -1,  348,   -1,  425,  276,   -1,   -1,   -1,  280,
   -1,   -1,   -1,   -1,  285,   -1,   -1,   -1,  289,   -1,
   -1,   -1,   -1,   -1,   -1,  296,   -1,  298,   -1,   -1,
  301,   -1,  303,   -1,  305,   -1,   -1,  308,   -1,  310,
  311,  312,  313,   -1,   -1,   -1,  317,  318,  319,   -1,
   -1,  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  331,  332,   -1,  334,  335,   -1,  337,  338,  339,   70,
   71,   72,   73,   74,   75,   76,   77,  348,   -1,  425,
   -1,  266,  267,  268,   -1,   -1,  271,  272,   -1,   -1,
  275,  276,   -1,   -1,  279,  280,  281,   -1,   -1,   -1,
   -1,  286,  287,   -1,  289,  290,   -1,  292,  293,   -1,
   -1,  296,   -1,   -1,   -1,  300,  301,   -1,  303,  304,
  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  315,   -1,  317,  261,  319,   -1,   -1,   -1,  323,   -1,
   -1,  326,  327,  328,  329,  273,  331,  332,  333,  334,
  335,   -1,   -1,  338,  425,  340,   -1,  285,   -1,   -1,
   -1,   -1,  347,  348,   -1,   -1,  351,   -1,   -1,   -1,
  298,  356,   -1,   -1,  302,  303,   -1,  178,  179,  180,
  308,   -1,  310,  311,  312,  313,  371,   -1,  373,   -1,
  318,   -1,   -1,   -1,  322,   -1,  348,   -1,  350,   -1,
  352,   -1,  354,  355,  356,   -1,  334,   -1,   -1,  337,
   -1,  339,  364,  365,   -1,   -1,  217,  369,  370,   -1,
  348,  349,   -1,   -1,   -1,   -1,   -1,   -1,  413,   -1,
  415,  383,  417,  385,  419,  387,  421,  389,  423,   -1,
  425,  266,  267,  268,   -1,   -1,  271,  272,   -1,   -1,
  275,  276,   -1,   -1,  279,  280,   -1,   -1,   -1,  262,
   -1,  286,   -1,   -1,  289,  290,   -1,  292,  293,   -1,
   -1,  296,   -1,   -1,   -1,  300,  301,   -1,  303,  304,
  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  315,   -1,  317,   -1,  319,   -1,  299,   -1,  323,   -1,
   -1,  326,  327,  328,  329,   -1,  331,  332,  333,  334,
  335,   -1,   -1,  338,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,  347,  348,   -1,   -1,  351,   -1,   -1,   -1,
   -1,  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  371,  350,  373,  352,
   -1,  354,  355,  356,   -1,  358,  359,   -1,   -1,  362,
  363,  364,  365,  366,  367,  368,  369,  370,   -1,   -1,
   -1,   -1,  375,   -1,  377,   -1,  379,   -1,  381,   -1,
  383,   -1,  385,   -1,  387,   -1,  389,   -1,  413,   -1,
  415,   -1,  417,   -1,  419,   -1,  421,   -1,  423,   -1,
  425,  266,  267,  268,   -1,   -1,  271,  272,   -1,   -1,
  275,  276,   -1,   -1,  279,  280,   -1,   -1,   -1,   -1,
   -1,  286,   -1,   -1,  289,  290,   -1,  292,  293,   -1,
   -1,  296,   -1,   -1,   -1,  300,  301,   -1,  303,  304,
  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  315,   -1,  317,   -1,  319,   -1,   -1,   -1,  323,   -1,
   -1,  326,  327,  328,  329,   -1,  331,  332,  333,  334,
  335,   -1,   -1,  338,  266,  340,  268,   -1,   -1,  271,
   -1,   -1,  347,   -1,  276,   -1,  351,   -1,  280,   -1,
   -1,  356,   -1,   -1,  286,   -1,   -1,  289,   -1,   -1,
   -1,   -1,   -1,   -1,  296,   -1,  371,   -1,  373,  301,
   -1,  303,  304,  305,   -1,   -1,  348,   -1,  350,   -1,
  352,   -1,  354,  355,  356,  317,   -1,  319,   -1,   -1,
   -1,  323,  364,  365,  326,   -1,  328,  369,  370,  331,
  332,   -1,   -1,  335,   -1,   -1,  338,   -1,  413,   -1,
  415,  383,  417,  385,  419,  387,  421,  389,  423,  351,
  425,   -1,   -1,   -1,   -1,  357,  358,  359,  360,   -1,
   -1,   -1,  364,   -1,  366,  266,   -1,  268,   -1,  371,
  271,  373,   -1,   -1,   -1,  276,   -1,   -1,   -1,  280,
   -1,   -1,   -1,   -1,   -1,  286,   -1,   -1,  289,   -1,
   -1,   -1,   -1,   -1,   -1,  296,   -1,   -1,   -1,   -1,
  301,   -1,  303,  304,  305,   -1,   -1,   -1,   -1,   -1,
   -1,  413,   -1,  415,   -1,  417,  317,  419,  319,  421,
   -1,  423,  323,  425,   -1,  326,   -1,  328,   -1,   -1,
  331,  332,   -1,   -1,  335,   -1,   -1,  338,  266,   -1,
  268,   -1,   -1,  271,   -1,   -1,   -1,   -1,  276,   -1,
  351,   -1,  280,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  289,   -1,   -1,   -1,   -1,   -1,   -1,  296,   -1,
  371,   -1,  373,  301,   -1,   -1,   -1,  305,  266,  307,
  268,   -1,   -1,  271,   -1,  273,  314,   -1,  276,  317,
   -1,  319,  280,   -1,   -1,  323,   -1,   -1,   -1,   -1,
   -1,  289,   -1,  331,  332,   -1,   -1,  335,  296,   -1,
  338,   -1,  413,  301,  415,   -1,  417,  305,  419,   -1,
  421,   -1,  423,  266,  425,  268,   -1,   -1,  271,  317,
   -1,  319,   -1,  276,   -1,  323,   -1,  280,   -1,   -1,
   -1,   -1,   -1,  331,  332,   -1,  289,  335,   -1,   -1,
  338,   -1,   -1,  296,   -1,  266,   -1,  268,  301,   -1,
  271,   -1,  305,   -1,   -1,  276,   -1,   -1,   -1,  280,
   -1,   -1,   -1,   -1,  317,   -1,  319,   -1,  289,   -1,
  323,   -1,   -1,   -1,   -1,  296,   -1,   -1,  331,  332,
  301,   -1,  335,   -1,  305,  338,   -1,  425,   -1,  266,
   -1,  268,   -1,   -1,  271,   -1,  317,   -1,  319,  276,
   -1,   -1,  323,  280,   -1,   -1,   -1,   -1,   -1,   -1,
  331,  332,  289,   -1,  335,   -1,   -1,  338,   -1,  296,
   -1,   -1,   -1,   -1,  301,   -1,   -1,  425,  305,  266,
   -1,  268,   -1,   -1,  271,   -1,   -1,   -1,   -1,  276,
  317,   -1,  319,  280,   -1,   -1,  323,   -1,  267,   -1,
   -1,   -1,  289,  272,  331,  332,  275,   -1,  335,  296,
  279,  338,   -1,   -1,  301,   -1,   -1,  286,  305,   -1,
   -1,  290,  425,  292,  293,   -1,  262,   -1,   -1,   -1,
  317,  300,  319,   -1,  303,  304,  323,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  331,  332,  315,   -1,  335,   -1,
   -1,  338,   -1,   -1,  425,   -1,   -1,  326,  327,  328,
  329,   -1,   -1,  299,  333,  334,   -1,   -1,  348,   -1,
  350,  340,  352,   -1,  354,  355,  356,   -1,  347,   -1,
   -1,   -1,  351,   -1,  364,  365,   -1,  356,   -1,  369,
  370,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  425,  262,
   -1,   -1,  371,  383,  373,  385,   -1,  387,   -1,  389,
   -1,   -1,  348,  349,  350,   -1,  352,   -1,  354,  355,
  356,   -1,  358,  359,   -1,  361,  362,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  299,  373,  425,  375,
   -1,  377,   -1,  379,  413,  381,  415,  383,  417,  385,
  419,  387,  421,  389,  423,  391,  425,  393,   -1,  395,
   -1,  397,   -1,  399,   -1,  401,   -1,  403,   -1,  405,
   -1,  407,  262,  409,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,   -1,  352,
   -1,  354,  355,  356,   -1,  358,  359,   -1,  361,  362,
  363,  364,  365,  366,  367,  368,  369,  370,  371,  299,
  373,   -1,  375,   -1,  377,   -1,  379,   -1,  381,   -1,
  383,   -1,  385,   -1,  387,   -1,  389,   -1,  391,   -1,
  393,   -1,  395,   -1,  397,   -1,  399,   -1,  401,   -1,
  403,   -1,  405,   -1,  407,  262,  409,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,   -1,  352,   -1,  354,  355,  356,   -1,  358,  359,
   -1,  361,  362,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  299,  373,   -1,  375,   -1,  377,   -1,  379,
   -1,  381,   -1,  383,   -1,  385,   -1,  387,   -1,  389,
   -1,  391,   -1,  393,   -1,  395,   -1,  397,   -1,  399,
   -1,  401,   -1,  403,   -1,  405,   -1,  407,  262,  409,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  347,  348,   -1,  350,   -1,  352,   -1,  354,  355,  356,
   -1,  358,  359,   -1,  361,  362,  363,  364,  365,  366,
  367,  368,  369,  370,  371,  299,  373,   -1,  375,   -1,
  377,   -1,  379,   -1,  381,   -1,  383,   -1,  385,   -1,
  387,   -1,  389,   -1,  391,   -1,  393,   -1,  395,  262,
  397,   -1,  399,   -1,  401,   -1,  403,   -1,  405,   -1,
  407,   -1,  409,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,   -1,  352,   -1,
  354,  355,  356,   -1,  358,  359,  299,  361,  362,  363,
  364,  365,  366,  367,  368,  369,  370,  371,   -1,  373,
   -1,  375,   -1,  377,   -1,  379,   -1,  381,   -1,  383,
   -1,  385,   -1,  387,   -1,  389,   -1,  391,   -1,  393,
   -1,  395,   -1,  397,   -1,  399,   -1,  401,   -1,  403,
   -1,  405,   -1,  407,   -1,  409,  349,   -1,  351,  352,
  353,  354,   -1,   -1,   -1,  358,  359,   -1,   -1,  362,
  363,  364,  365,  366,  367,  368,  369,  370,  371,   -1,
  373,   -1,  375,  261,  377,   -1,  379,   -1,  381,   -1,
  383,   -1,  385,   -1,  387,   -1,  389,   -1,  391,   -1,
  393,  286,  395,   -1,  397,   -1,  399,  285,  401,   -1,
  403,   -1,  405,   -1,  407,   -1,  409,   -1,  303,  304,
  298,   -1,  307,   -1,  302,  303,   -1,   -1,  286,  314,
  308,   -1,  310,  311,  312,  313,   -1,   -1,   -1,   -1,
  318,  326,   -1,  328,  322,  303,  304,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  334,   -1,   -1,  337,
   -1,  339,   -1,   -1,   -1,   -1,  351,   -1,  326,   -1,
  328,   -1,  357,  358,  359,  360,   -1,   -1,   -1,  364,
   -1,  366,   -1,   -1,   -1,   -1,  371,   -1,  373,  347,
  348,   -1,   -1,  351,   -1,   -1,   -1,   -1,   -1,  357,
  358,  359,  360,  286,   -1,   -1,  364,   -1,  366,   -1,
   -1,   -1,   -1,  371,   -1,  373,   -1,   -1,   -1,   -1,
  303,  304,   -1,   -1,  286,   -1,   -1,   -1,  413,   -1,
  415,   -1,  417,   -1,  419,   -1,  421,   -1,  423,   -1,
  425,  303,  304,  326,   -1,  328,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  413,   -1,  415,   -1,  417,
   -1,  419,   -1,  421,  326,  423,  328,  425,  351,   -1,
   -1,  354,   -1,   -1,  357,  358,  359,  360,   -1,   -1,
   -1,  364,   -1,  366,   -1,  347,   -1,   -1,  371,  351,
  373,   -1,   -1,   -1,   -1,  357,  358,  359,  360,  286,
   -1,   -1,  364,   -1,  366,   -1,  348,   -1,  350,  371,
  352,  373,  354,  355,  356,   -1,  303,  304,   -1,   -1,
   -1,   -1,  364,  365,   -1,   -1,  286,  369,  370,   -1,
  413,   -1,  415,   -1,  417,   -1,  419,   -1,  421,  326,
  423,  328,  425,  303,  304,  387,   -1,  389,   -1,   -1,
   -1,  413,   -1,  415,   -1,  417,   -1,  419,   -1,  421,
   -1,  423,   -1,  425,  351,   -1,  326,   -1,  328,  356,
  357,  358,  359,  360,   -1,   -1,   -1,  364,   -1,  366,
   -1,   -1,   -1,   -1,  371,   -1,  373,   -1,   -1,   -1,
   -1,  351,   -1,   -1,   -1,   -1,  356,  357,  358,  359,
  360,  286,   -1,   -1,  364,   -1,  366,   -1,   -1,   -1,
   -1,  371,   -1,  373,   -1,   -1,   -1,   -1,  303,  304,
   -1,   -1,  286,   -1,   -1,   -1,  413,  286,  415,   -1,
  417,   -1,  419,   -1,  421,   -1,  423,   -1,  425,  303,
  304,  326,   -1,  328,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  413,   -1,  415,   -1,  417,   -1,  419,
   -1,  421,  326,  423,  328,  425,  351,  326,   -1,  328,
   -1,   -1,  357,  358,  359,  360,   -1,   -1,   -1,  364,
   -1,  366,  262,   -1,   -1,   -1,  371,  351,  373,   -1,
   -1,   -1,  351,  357,  358,  359,  360,   -1,   -1,   -1,
  364,   -1,  366,   -1,   -1,   -1,   -1,  371,   -1,  373,
   -1,   -1,  371,   -1,  373,   -1,   -1,   -1,   -1,  299,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,  413,   -1,
  415,   -1,  417,   -1,  419,   -1,  421,   -1,  423,   -1,
  425,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  413,
   -1,  415,   -1,  417,  413,  419,  415,  421,  417,  423,
  419,  425,  421,  299,  423,   -1,  425,   -1,  348,  262,
  350,   -1,  352,   -1,  354,  355,  356,   -1,  358,  359,
   -1,   -1,  362,  363,  364,  365,   -1,   -1,   -1,  369,
  370,   -1,   -1,   -1,   -1,  375,   -1,  377,   -1,  379,
   -1,  381,   -1,  383,   -1,  385,  299,  387,   -1,  389,
   -1,  262,  348,   -1,  350,   -1,  352,   -1,  354,  355,
  356,   -1,  358,  359,   -1,   -1,  362,  363,  364,  365,
   -1,   -1,   -1,  369,  370,   -1,   -1,   -1,   -1,  375,
   -1,  377,   -1,  379,   -1,  381,   -1,  383,  299,  385,
   -1,  387,   -1,  389,  262,  348,   -1,  350,   -1,  352,
   -1,  354,  355,  356,   -1,  358,  359,   -1,   -1,  362,
  363,  364,  365,   -1,   -1,   -1,  369,  370,   -1,   -1,
   -1,   -1,  375,   -1,  377,   -1,  379,   -1,  381,   -1,
  383,  299,  385,   -1,  387,   -1,  389,  348,  262,  350,
   -1,  352,   -1,  354,  355,  356,   -1,   -1,   -1,   -1,
   -1,  362,  363,  364,  365,   -1,   -1,   -1,  369,  370,
   -1,   -1,   -1,   -1,  375,   -1,  377,   -1,  379,   -1,
  381,   -1,  383,   -1,  385,  299,  387,   -1,  389,   -1,
  348,  262,  350,   -1,  352,   -1,  354,  355,  356,   -1,
   -1,   -1,   -1,   -1,  362,  363,  364,  365,   -1,   -1,
   -1,  369,  370,   -1,   -1,   -1,   -1,  375,   -1,  377,
   -1,  379,   -1,  381,   -1,  383,   -1,  385,  299,  387,
   -1,  389,   -1,  262,  348,   -1,  350,   -1,  352,   -1,
  354,  355,  356,   -1,   -1,   -1,   -1,   -1,  362,  363,
  364,  365,   -1,   -1,   -1,  369,  370,   -1,   -1,   -1,
   -1,  375,   -1,  377,   -1,  379,   -1,  381,   -1,  383,
  299,  385,   -1,  387,   -1,  389,  262,  348,   -1,  350,
   -1,  352,   -1,  354,  355,  356,   -1,   -1,   -1,   -1,
   -1,  362,  363,  364,  365,   -1,   -1,   -1,  369,  370,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  379,   -1,
  381,   -1,  383,  299,  385,   -1,  387,   -1,  389,  348,
  262,  350,   -1,  352,   -1,  354,  355,  356,   -1,   -1,
   -1,   -1,   -1,  362,  363,  364,  365,   -1,   -1,   -1,
  369,  370,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  379,   -1,  381,   -1,  383,   -1,  385,  299,  387,   -1,
  389,   -1,  348,  262,  350,   -1,  352,   -1,  354,  355,
  356,   -1,   -1,   -1,   -1,   -1,  362,  363,  364,  365,
   -1,   -1,   -1,  369,  370,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  379,   -1,  381,   -1,  383,   -1,  385,
  299,  387,   -1,  389,   -1,  262,  348,   -1,  350,   -1,
  352,   -1,  354,  355,  356,   -1,   -1,   -1,   -1,   -1,
  362,  363,  364,  365,   -1,   -1,   -1,  369,  370,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  379,   -1,  381,
   -1,  383,  299,  385,   -1,  387,   -1,  389,  262,  348,
   -1,  350,   -1,  352,   -1,  354,  355,  356,   -1,   -1,
   -1,   -1,   -1,  362,  363,  364,  365,   -1,   -1,   -1,
  369,  370,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  379,   -1,  381,   -1,  383,  299,  385,   -1,  387,   -1,
  389,  348,   -1,  350,   -1,  352,   -1,  354,  355,  356,
   -1,   -1,   -1,   -1,   -1,  362,  363,  364,  365,   -1,
   -1,   -1,  369,  370,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  379,   -1,  381,   -1,  383,   -1,  385,   -1,
  387,   -1,  389,   -1,  348,  261,  350,   -1,  352,   -1,
  354,  355,  356,   -1,   -1,   -1,   -1,  273,  362,  363,
  364,  365,   -1,   -1,   -1,  369,  370,   -1,   -1,  285,
   -1,   -1,   -1,   -1,   -1,  379,   -1,  381,   -1,  383,
   -1,  385,  298,  387,   -1,  389,  302,  303,  261,   -1,
   -1,   -1,  308,   -1,  310,  311,  312,  313,   -1,   -1,
  273,   -1,  318,   -1,   -1,   -1,  322,   -1,   -1,   -1,
   -1,   -1,  285,   -1,   -1,   -1,   -1,   -1,  334,   -1,
   -1,  337,   -1,  339,   -1,  298,   -1,   -1,   -1,  302,
  303,  261,  348,  349,   -1,  308,   -1,  310,  311,  312,
  313,   -1,   -1,  273,   -1,  318,   -1,   -1,   -1,  322,
   -1,   -1,   -1,   -1,   -1,  285,   -1,   -1,   -1,   -1,
   -1,  334,   -1,   -1,  337,   -1,  339,   -1,  298,   -1,
   -1,   -1,  302,  303,   -1,  348,   -1,   -1,  308,   -1,
  310,  311,  312,  313,  349,   -1,   -1,  352,  318,  354,
   -1,  356,  322,   -1,   -1,   -1,  361,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  371,  337,  373,  339,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  349,  348,   -1,
  352,   -1,  354,   -1,  356,   -1,  391,   -1,  393,  361,
  395,   -1,  397,   -1,  399,   -1,  401,   -1,  403,  371,
  405,  373,  407,   -1,  409,   -1,  349,   -1,   -1,  352,
   -1,  354,   -1,  356,   -1,   -1,   -1,   -1,  361,  391,
   -1,  393,   -1,  395,   -1,  397,   -1,  399,  371,  401,
  373,  403,   -1,  405,   -1,  407,   -1,  409,   -1,  349,
   -1,   -1,  352,   -1,  354,   -1,  356,   -1,  391,   -1,
  393,  361,  395,   -1,  397,   -1,  399,   -1,  401,   -1,
  403,  371,  405,  373,  407,   -1,  409,   -1,  349,   -1,
  351,   -1,  353,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  361,  391,   -1,  393,   -1,  395,   -1,  397,   -1,  399,
  371,  401,  373,  403,   -1,  405,   -1,  407,   -1,  409,
   -1,  349,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  391,   -1,  393,  361,  395,   -1,  397,   -1,  399,   -1,
  401,   -1,  403,  371,  405,  373,  407,   -1,  409,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  391,   -1,  393,   -1,  395,   -1,  397,
   -1,  399,   -1,  401,   -1,  403,  348,  405,  350,  407,
  352,  409,  354,  355,  356,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  364,  365,   -1,   -1,   -1,  369,  370,  348,
   -1,  350,   -1,  352,   -1,  354,  355,  356,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  387,  365,  389,   -1,   -1,
  369,  370,  348,   -1,  350,   -1,  352,   -1,  354,  355,
  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  387,  365,
  389,   -1,   -1,  369,  370,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  387,   -1,  389,
  };

#line 1620 "cs-parser.jay"

    //
    // Helper functions
    //


    /**
     * Resolve given reference to a VariableDecl (block variable, method parameter,
     * class field). We cannot use Ref.resolve() for that since there is not
     * yet a method call stack, nor a Class.getCurrentClass(). This also checks
     * if an ID is allowed at the calling place at all (i. e. field
     * initializers must be constant expressions), and that non-stub QIDs must
     * be unqualified (restriction in cs2ccmb since we only deal with one
     * class).
     */
    VariableDecl resolveVar( Ref r ) {
	VariableDecl v = null;
	if( !r.isThis() ) {
	    v = currBlock.getVar( r.getId().ToString() );
	    if( v == null ) 
		v = currMethod.getParam( r.getId().ToString() );
	}
	if( v == null )
	    v = currClass.getField( r.getId().id );
	if( v == null ) {
	    errUndeclaredId( r.ToString() );
	    return null;
	} 

	if( v.type != stubType && r.isQualified() ) {
	    errNonStubMember( r.getId().id );
	    return null;
	}

	if( r.getId().next != null ) {
	    v = v.stubClass.getField( r.getId().next.id );
	    if( v == null ) {
		errUndeclaredId( r.ToString() );
		return null;
	    }
	}

	return v;
    }

    /// Resolve given reference to a Method.
    Method resolveMethod( Ref r, IList args ) {
	string first = r.getId().id;

	Method m = null;
	if( !r.isQualified() ) {
	    m = currClass.getMethod( first, args );
	    if( m != null )
		return m;
	    errUndeclaredId( first );
	    return null;
	}

	string second = r.getId().next.id;
	if( r.getId().next.next != null ) {
	    errUndeclaredId( r.ToString() );
	    return null;
	}

	// qualified; check if it is a class (static method)
	Class c = (Class) classes[first];
	if( c != null ) {
	    m = c.getMethod( second, args );
	    if( m == null ) {
		errUndeclaredId( r.ToString() );
		return null;
	    }
	    if( !m.isStatic() ) {
		errNonStaticMethod();
		return null;
	    }
	    return m;
	}

	// non static; resolve stub and search method in it
	VariableDecl v = resolveVar( new Ref( new QID( first ), r.isThis() ) );
	if( v.type != stubType ) {
	    errNonStubMember( first );
	    return null;
	}
	m = v.stubClass.getMethod( second, args );
	if( m != null )
	    return m;
	errUndeclaredId( r.ToString() );
	return null;
    }

    /**
     * Create a reference expression to a given variable declaration.
     * Note that we cannot directly save v into a Ref since at runtime the
     * IDs will resolve to different variables.
     */
    Expr makeRefExpr( VariableDecl v, Ref r ) {
	if( v.array )
	    return new ArrayExpr( r, v.type );
	if( v.type == false.GetType() )
	    return new BoolExpr( r );
	if( v.type == 0.GetType() )
	    return new IntExpr( r );
	if( v.type == (0.0d).GetType() )
	    return new FloatExpr( r );
	if( v.type == "".GetType() )
	    return new StringExpr( r );
	if( v.type == stubType )
	    return new StubExpr( r );
	throw new Exception( "Internal error: unknown type " + v.type );
    }

    /**
     * Create and return an assignment expression for a = e. 
     * @param a assignable expression; it is just an object, so that invocations from the parser
     *        rules can be written without a typecast
     * @param e expression which is assigned to a
     */
    Expr makeAssign( Object a, Expr e ) {
	if( e == null ) // happens on syntax error, exit cleanly
	    return null;

	Expr asgn = a as Expr;
	if( asgn == null ) 
	    throw new Exception( lexer.Location + " CSharpParser.makeAssign(): parameter is no Expression" );
	if( !asgn.assignable() ) {
	    errNotAssignable( asgn );
	    return null;
	}

	// array element assignments must be treated specially because Refs don't
	// save indices; maybe there's a better architecture?
	if( asgn.isElementAccess() ) {
	    Type t = asgn.type();
	    if( t != e.type() ) {
		errAssignmentTypeMismatch();
		return null;
	    }
	    return asgn.makeElementAssign( e );
	}

	Ref r = asgn.getRef();

	return makeRefAssign( r, e );
    }

    /**
     * Return an assignment Expression for r = e. Needed for local
     * variable initializers, where there is no left-hand expression.
     */
    Expr makeRefAssign( Ref r, Expr e ) {
	VariableDecl v = resolveVar( r );

	if( v.array ) {
	    if( !( e is NullExpr ) && v.type != ( (ArrayExpr) e).elementType() ) {
		errAssignmentTypeMismatch();
		return null;
	    } 
	    return new ArrayExpr( r, e, v.type );
	}

	if( !( e is NullExpr ) && v.type != e.type() ) {
	    errAssignmentTypeMismatch();
	    return null;
	}

	if( v.type == false.GetType() )
	    return new BoolExpr( r, e );
	if( v.type == 0.GetType() )
	    return new IntExpr( r, e );
	if( v.type == (0.0d).GetType() )
	    return new FloatExpr( r, e );
	if( v.type == "".GetType() )
	    return new StringExpr( r, e );
	if( v.type == stubType )
	    return new StubExpr( r, e );

	throw new Exception( "CSharpParser.makeRefAssign(): invalid expression type " + e.GetType() );
    }

    /// Return a method call expression for m(pars).
    Expr makeCall( Ref r, Method m, List pars ) {
	Type t = m.returnType();
	if( !m.compatibleArgs( pars ) ) {
	    errWrongCallParams(); 
	    return null;
	}

	if( t == null )
	    return new VoidExpr( r, pars );
	if( t == false.GetType() )
	    return new BoolExpr( r, pars );
	if( t == 0.GetType() )
	    return new IntExpr( r, pars );
	if( t == (0.0d).GetType() )
	    return new FloatExpr( r, pars );
	if( t == "".GetType() )
	    return new StringExpr( r, pars );

	throw new Exception( "Internal error: unknown type " + t );
    }

    //
    // Type checking functions
    //

    /// Assert that given object is a boolean expression and report error if not.
    void chkb( Object o ) {
	if( !( o is BoolExpr ) )
	    Report.Error( 1, lexer.Location, "boolean type expression expected" );
    }

    /// Assert that given object is an integer expression and report error if not.
    void chki( Object o ) {
	if( !( o is IntExpr ) )
	    Report.Error( 2, lexer.Location, "integer type expression expected" );
    }

    /// Assert that given object is a double expression and report error if not.
    void chkf( Object o ) {
	if( !( o is FloatExpr ) )
	    Report.Error( 3, lexer.Location, "float type expression expected" );
    }

    /// Assert that given object is a string expression and report error if not.
    void chks( Object o ) {
	 if( !( o is StringExpr ) )
	    Report.Error( 4, lexer.Location, "string type expression expected" );
    }

    /// Assert that given objects are expressions of the same type and report
    /// an error if not.
    void chksametype( Object o1, Object o2 ) {
	if( o1.GetType() != o2.GetType() )
	    Report.Error( 5, lexer.Location, "types must be equal for comparison" );
    }	

    //
    // Error reporting functions
    //

    /// Error report: numeric expression expected.
    void errNumExp() {
	Report.Error( 6, lexer.Location, "numeric type (int or double) expected" );
    }

    /// Error report: assignment type mismatch.
    void errAssignmentTypeMismatch() {
	Report.Error( 7, lexer.Location, "types must match for assignment" );
    }

    /// Error report: expression is not assignable.
    void errNotAssignable( Expr e ) {
	Report.Error( 8, lexer.Location, "expression " + e + " is not assignable" );
    }

    /// Error report: unknown type.
    void errUnknownType() {
	Report.Error( 9, lexer.Location, "unknown type");
    }

    /// Error report: break or continue used outside of a loop.
    void errBreakContNoLoop() {
	Report.Error( 100, lexer.Location, "break and continue may only be used in loops" );
    }

    /// Error report: undeclared identifier.
    void errUndeclaredId( string id ) {
	Report.Error( 101, lexer.Location, "undeclared identifier: "+id );
    }

    /// Error report: return without value in non-void method.
    void errVoidReturn() {
	Report.Error( 102, lexer.Location, "return must have a value in non-void methods" );
    }

    /// Error report: return with value in void method.
    void errNoVoidReturn() {
	Report.Error( 103, lexer.Location, "return must not have a value in void methods" );
    }

    /// Error report: return value does not match method return type.
    void errReturnType() {
	Report.Error( 104, lexer.Location, "return value type must match method return type" );
    }

    /// Error report: incompatible arguments in method call.
    void errWrongCallParams() {
	Report.Error( 105, lexer.Location, "incompatible arguments in method call" );
    }

    /// Error report: new must be used with class types.
    void errNewType() {
	Report.Error( 108, lexer.Location, "operator 'new' must be used with class types" );
    }

    /// Error report: syntax error in new expression.
    void errNewSyntax() {
        Report.Error ( 109, lexer.Location, "new expression requires () or [] after type" ); 
    }

    /// Error report: element access of non-array expression.
    void errNonArrayElementAccess( string e ) {
	Report.Error( 110, lexer.Location, e + " is not an array, cannot be [] accessed" );
    }

    /// Error report: no matching constructor for given arguments.
    void errNoMatchingCtor() {
	Report.Error( 111, lexer.Location, "did not find a matching constructor" );
    }

    /// Error report: static constructor is forbidden.
    void errStaticCtor() {
	Report.Error( 112, lexer.Location, "constructors must not be static" );
    }

    /// Error report: nonstatic method called of static class reference.
    void errNonStaticMethod() {
	Report.Error( 113, lexer.Location, "nonstatic method called from class" );
    }

    /// Error report: ambiguous method declaration.
    void errDuplicateMethod() {
	Report.Error( 114, lexer.Location, "A method with this name and parameters already exists" );
    }

    /// Error report: unimplemented operator.
    void errUnimplOp() {
	Report.Error( 1000, lexer.Location, "unimplemented operator" );
    }

    /// Error report: invalid class ancestor.
    void errClassBase() {
	Report.Error( 1001, lexer.Location, "classes must be derived from Stub, StubR, StubS, or Comp" );
    }

    /// Error report: stubs are invalid as local block variables.
    void errStubTypeForbidden() {
	Report.Error( 1002, lexer.Location, "stub types are only valid as class fields" );
    }

    /// Error report: member access of non-stub expression.
    void errNonStubMember( string var ) {
	Report.Error( 1003, lexer.Location, var + " is no Stub, no member access allowed" );
    }

    /// Error report: currently supports only one-dimensional arrays.
    void errOneDimArrayAccess() {
	Report.Error( 1004, lexer.Location, "currently only singe-dimension arrays are supported" );
    }

    /// C# scanner (lexical analysis).
    Tokenizer lexer;

    /**
     * Parse a C# module. Classes are added to the list given in the
     * constructor. 
     * @param reader input TextReader that contains file to be parsed
     * @param fname file name of reader to give proper error messages
     * @return true on success or false if given reader cannot be parsed as C#
     *         module
     */
    public bool parse( TextReader reader, string fname )
    {
	lexer = new Tokenizer( reader, fname, null );
	try {
	    yyparse( lexer /*, new yydebug.yyDebugSimple() */ );
	    return Report.Errors == 0;
	} catch( Exception e ) {
	    Console.WriteLine( "Exception caught in " + lexer.location + ":" );
	    Console.WriteLine( e );
	    return false;
	}
    }

}
#line default
namespace yydebug {
        using System;
	 public interface yyDebug {
		 void push (int state, Object value);
		 void lex (int state, int token, string name, Object value);
		 void shift (int from, int to, int errorFlag);
		 void pop (int state);
		 void discard (int state, int token, string name, Object value);
		 void reduce (int from, int to, int rule, string text, int len);
		 void shift (int from, int to);
		 void accept (Object value);
		 void error (string message);
		 void reject ();
	 }
	 
	 class yyDebugSimple : yyDebug {
		 void println (string s){
			 Console.WriteLine (s);
		 }
		 
		 public void push (int state, Object value) {
			 println ("push\tstate "+state+"\tvalue "+value);
		 }
		 
		 public void lex (int state, int token, string name, Object value) {
			 println("lex\tstate "+state+"\treading "+name+"\tvalue "+value);
		 }
		 
		 public void shift (int from, int to, int errorFlag) {
			 switch (errorFlag) {
			 default:				// normally
				 println("shift\tfrom state "+from+" to "+to);
				 break;
			 case 0: case 1: case 2:		// in error recovery
				 println("shift\tfrom state "+from+" to "+to
					     +"\t"+errorFlag+" left to recover");
				 break;
			 case 3:				// normally
				 println("shift\tfrom state "+from+" to "+to+"\ton error");
				 break;
			 }
		 }
		 
		 public void pop (int state) {
			 println("pop\tstate "+state+"\ton error");
		 }
		 
		 public void discard (int state, int token, string name, Object value) {
			 println("discard\tstate "+state+"\ttoken "+name+"\tvalue "+value);
		 }
		 
		 public void reduce (int from, int to, int rule, string text, int len) {
			 println("reduce\tstate "+from+"\tuncover "+to
				     +"\trule ("+rule+") "+text);
		 }
		 
		 public void shift (int from, int to) {
			 println("goto\tfrom state "+from+" to "+to);
		 }
		 
		 public void accept (Object value) {
			 println("accept\tvalue "+value);
		 }
		 
		 public void error (string message) {
			 println("error\t"+message);
		 }
		 
		 public void reject () {
			 println("reject");
		 }
		 
	 }
}
// %token constants
 class Token {
  public const int EOF = 257;
  public const int NONE = 258;
  public const int ERROR = 259;
  public const int FIRST_KEYWORD = 260;
  public const int ABSTRACT = 261;
  public const int AS = 262;
  public const int ADD = 263;
  public const int ASSEMBLY = 264;
  public const int BASE = 265;
  public const int BOOL = 266;
  public const int BREAK = 267;
  public const int BYTE = 268;
  public const int CASE = 269;
  public const int CATCH = 270;
  public const int CHAR = 271;
  public const int CHECKED = 272;
  public const int CLASS = 273;
  public const int CONST = 274;
  public const int CONTINUE = 275;
  public const int DECIMAL = 276;
  public const int DEFAULT = 277;
  public const int DELEGATE = 278;
  public const int DO = 279;
  public const int DOUBLE = 280;
  public const int ELSE = 281;
  public const int ENUM = 282;
  public const int EVENT = 283;
  public const int EXPLICIT = 284;
  public const int EXTERN = 285;
  public const int FALSE = 286;
  public const int FINALLY = 287;
  public const int FIXED = 288;
  public const int FLOAT = 289;
  public const int FOR = 290;
  public const int FOREACH = 291;
  public const int GOTO = 292;
  public const int IF = 293;
  public const int IMPLICIT = 294;
  public const int IN = 295;
  public const int INT = 296;
  public const int INTERFACE = 297;
  public const int INTERNAL = 298;
  public const int IS = 299;
  public const int LOCK = 300;
  public const int LONG = 301;
  public const int NAMESPACE = 302;
  public const int NEW = 303;
  public const int NULL = 304;
  public const int OBJECT = 305;
  public const int OPERATOR = 306;
  public const int OUT = 307;
  public const int OVERRIDE = 308;
  public const int PARAMS = 309;
  public const int PRIVATE = 310;
  public const int PROTECTED = 311;
  public const int PUBLIC = 312;
  public const int READONLY = 313;
  public const int REF = 314;
  public const int RETURN = 315;
  public const int REMOVE = 316;
  public const int SBYTE = 317;
  public const int SEALED = 318;
  public const int SHORT = 319;
  public const int SIZEOF = 320;
  public const int STACKALLOC = 321;
  public const int STATIC = 322;
  public const int STRING = 323;
  public const int STRUCT = 324;
  public const int SWITCH = 325;
  public const int THIS = 326;
  public const int THROW = 327;
  public const int TRUE = 328;
  public const int TRY = 329;
  public const int TYPEOF = 330;
  public const int UINT = 331;
  public const int ULONG = 332;
  public const int UNCHECKED = 333;
  public const int UNSAFE = 334;
  public const int USHORT = 335;
  public const int USING = 336;
  public const int VIRTUAL = 337;
  public const int VOID = 338;
  public const int VOLATILE = 339;
  public const int WHILE = 340;
  public const int YIELD = 341;
  public const int GET = 342;
  public const int get = 343;
  public const int SET = 344;
  public const int set = 345;
  public const int LAST_KEYWORD = 346;
  public const int OPEN_BRACE = 347;
  public const int CLOSE_BRACE = 348;
  public const int OPEN_BRACKET = 349;
  public const int CLOSE_BRACKET = 350;
  public const int OPEN_PARENS = 351;
  public const int CLOSE_PARENS = 352;
  public const int DOT = 353;
  public const int COMMA = 354;
  public const int COLON = 355;
  public const int SEMICOLON = 356;
  public const int TILDE = 357;
  public const int PLUS = 358;
  public const int MINUS = 359;
  public const int BANG = 360;
  public const int ASSIGN = 361;
  public const int OP_LT = 362;
  public const int OP_GT = 363;
  public const int BITWISE_AND = 364;
  public const int BITWISE_OR = 365;
  public const int STAR = 366;
  public const int PERCENT = 367;
  public const int DIV = 368;
  public const int CARRET = 369;
  public const int INTERR = 370;
  public const int OP_INC = 371;
  public const int OP_DEC = 373;
  public const int OP_SHIFT_LEFT = 375;
  public const int OP_SHIFT_RIGHT = 377;
  public const int OP_LE = 379;
  public const int OP_GE = 381;
  public const int OP_EQ = 383;
  public const int OP_NE = 385;
  public const int OP_AND = 387;
  public const int OP_OR = 389;
  public const int OP_MULT_ASSIGN = 391;
  public const int OP_DIV_ASSIGN = 393;
  public const int OP_MOD_ASSIGN = 395;
  public const int OP_ADD_ASSIGN = 397;
  public const int OP_SUB_ASSIGN = 399;
  public const int OP_SHIFT_LEFT_ASSIGN = 401;
  public const int OP_SHIFT_RIGHT_ASSIGN = 403;
  public const int OP_AND_ASSIGN = 405;
  public const int OP_XOR_ASSIGN = 407;
  public const int OP_OR_ASSIGN = 409;
  public const int OP_PTR = 411;
  public const int LITERAL_INTEGER = 413;
  public const int LITERAL_FLOAT = 415;
  public const int LITERAL_DOUBLE = 417;
  public const int LITERAL_DECIMAL = 419;
  public const int LITERAL_CHARACTER = 421;
  public const int LITERAL_STRING = 423;
  public const int IDENTIFIER = 425;
  public const int CLOSE_PARENS_CAST = 426;
  public const int CLOSE_PARENS_NO_CAST = 427;
  public const int CLOSE_PARENS_OPEN_PARENS = 428;
  public const int CLOSE_PARENS_MINUS = 429;
  public const int LOWPREC = 430;
  public const int UMINUS = 431;
  public const int HIGHPREC = 432;
  public const int yyErrorCode = 256;
 }
 namespace yyParser {
  using System;
  /** thrown for irrecoverable syntax errors and stack overflow.
    */
  public class yyException : System.Exception {
    public yyException (string message) : base (message) {
    }
  }

  /** must be implemented by a scanner object to supply input to the parser.
    */
  public interface yyInput {
    /** move on to next token.
        @return false if positioned beyond tokens.
        @throws IOException on input error.
      */
    bool advance (); // throws java.io.IOException;
    /** classifies current token.
        Should not be called if advance() returned false.
        @return current %token or single character.
      */
    int token ();
    /** associated with current token.
        Should not be called if advance() returned false.
        @return value for token().
      */
    Object value ();
  }
 }
} // close outermost namespace, that MUST HAVE BEEN opened in the prolog
